# -------------------------------------------------------------------------
#
# Author:      RRD
#
# Created:     24/10/2012
# Copyright:   (c) rdamiani 2012
# Licence:     <your licence>
# -------------------------------------------------------------------------


import numpy as np


def frustum(Db, Dt, H):
    """This function returns a frustum's volume and center of mass, CM

    INPUT:
    Parameters
    ----------
    Db : float,        base diameter
    Dt : float,        top diameter
    H : float,         height

    OUTPUTs:
    -------
    vol : float,        volume
    cm : float,        geometric centroid relative to bottom (center of mass if uniform density)

    """
    vol = frustumVol(Db, Dt, H, diamFlag=True)
    cm = frustumCG(Db, Dt, H, diamFlag=True)
    # vol = np.pi/12*H * (Db**2 + Dt**2 + Db * Dt)
    # cm = H/4 * (Db**2 + 3*Dt**2 + 2*Db*Dt) / (Db**2 + Dt**2 + Db*Dt)
    return vol, cm


def frustumVol(rb_0, rt_0, h, diamFlag=False):
    """This function returns a frustum's volume with radii or diameter inputs.

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    vol : float (scalar/vector), volume
    """
    if diamFlag:
        # Convert diameters to radii
        rb, rt = 0.5 * rb_0, 0.5 * rt_0
    else:
        rb, rt = rb_0, rt_0
    return np.pi * (h / 3.0) * (rb * rb + rt * rt + rb * rt)

def RectangularFrustumVol(ab, bb, at, bt, h):
    """This function returns a rectangular frustum's volume with side length inputs.

    INPUTS:
    Parameters
    ----------
    ab : float (scalar/vector),  base side length a
    bb : float (scalar/vector),  base side length b
    at : float (scalar/vector),  top side length a
    bt : float (scalar/vector),  top side length b
    h  : float (scalar/vector),  height

    OUTPUTs:
    -------
    vol : float (scalar/vector), volume
    """

    Abottom = ab * bb
    Atop = at * bt

    V = GeneralFrustumVol(Ab=Abottom, At=Atop, h=h)
    
    return V

def GeneralFrustumVol(Ab, At, h):
    """This function returns a frustum's volume with top and base areas, and height

    INPUTS:
    Parameters
    ----------
    Ab : float (scalar/vector),  base area
    At : float (scalar/vector),  top area
    h  : float (scalar/vector),  height

    OUTPUTs:
    -------
    vol : float (scalar/vector), volume
    """

    Atop = At
    Abottom = Ab

    V = h / 3.0 * (Atop + Abottom + np.sqrt(Atop * Abottom))
    
    return V

def frustumCG(rb_0, rt_0, h, diamFlag=False):
    """This function returns a frustum's center of mass/gravity (centroid) with radii or diameter inputs.
    NOTE: This is for a SOLID frustum, not a shell

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    cg : float (scalar/vector),  center of mass/gravity (centroid)
    """
    if diamFlag:
        # Convert diameters to radii
        rb, rt = 0.5 * rb_0, 0.5 * rt_0
    else:
        rb, rt = rb_0, rt_0
    return 0.25 * h * (rb**2 + 2.0 * rb * rt + 3.0 * rt**2) / (rb**2 + rb * rt + rt**2)

def RectangularFrustumCG(ab, bb, at, bt, h):
    """This function returns a rectangular frustum's center of mass/gravity (centroid) with side lengths and height
    NOTE: This is for a SOLID frustum, not a shell

    INPUTS:
    Parameters
    ----------
    ab : float (scalar/vector),  base side length a
    bb : float (scalar/vector),  base side length b
    at : float (scalar/vector),  top side length a
    bt : float (scalar/vector),  top side length b
    h  : float (scalar/vector),  height

    OUTPUTs:
    -------
    cg : float (scalar/vector),  center of mass/gravity (centroid)
    """
    cg = GeneralFrustumCG(Ab = ab*bb, At = at*bt, h = h)

    return cg


def GeneralFrustumCG(Ab, At, h):
    """This function returns a frustum's center of mass/gravity (centroid) with base and top area inputs.
    NOTE: This is for a SOLID frustum, not a shell

    INPUTS:
    Parameters
    ----------
    Ab : float (scalar/vector),  base area
    At : float (scalar/vector),  top area
    h  : float (scalar/vector),  height
    V  : float (scalar/vector),  Volume

    OUTPUTs:
    -------
    cg : float (scalar/vector),  center of mass/gravity (centroid)
    """
    cg = h * (Ab + 2*np.sqrt(Ab*At) + 3*At) / (4 * (Ab + np.sqrt(Ab * At) + At))

    return cg

def frustumIzz(rb_0, rt_0, h, diamFlag=False):
    """This function returns a frustum's mass-moment of inertia (divided by density) about the
    central (axial) z-axis with radii or diameter inputs.
    NOTE: This is for a SOLID frustum, not a shell

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    Izz : float (scalar/vector),  Moment of inertia about z-axis
    """
    if diamFlag:
        # Convert diameters to radii
        rb, rt = 0.5 * rb_0, 0.5 * rt_0
    else:
        rb, rt = rb_0, rt_0
    # Integrate 2*pi*r*r^2 dr dz from r=0 to r(z), z=0 to h
    # Also equals 0.3*Vol * (rt**5.0 - rb**5.0) / (rt**3.0 - rb**3.0)
    # Also equals (0.1*np.pi*h * (rt**5.0 - rb**5.0) / (rt - rb) )
    return 0.1 * np.pi * h * (rt**4.0 + rb * rt**3 + rb**2 * rt**2 + rb**3 * rt + rb**4.0)

def RectangularFrustumIzz(ab, bb, at, bt, h):
    """This function returns a rectangular frustum's mass-moment of inertia (divided by density) about the
    central (axial) z-axis with side lengths and height
    NOTE: This is for a SOLID frustum, not a shell

    INPUTS:
    Parameters
    ----------
    ab : float (scalar/vector),  base side length a
    bb : float (scalar/vector),  base side length b
    at : float (scalar/vector),  top side length a
    bt : float (scalar/vector),  top side length b
    h  : float (scalar/vector),  height

    OUTPUTs:
    -------
    Izz : float (scalar/vector),  Moment of inertia about z-axis
    """

    # Use calculations from RAFT

    x2 = (1/12) * ( (at-ab)**3*h*(bt/5 + bb/20) + (at-ab)**2*ab*h*(3*bt/4 + bb/4) + \
                    (at-ab)*ab**2*h*(bt + bb/2) + ab**3*h*(bt/2 + bb/2) )

    y2 = (1/12) * ( (bt-bb)**3*h*(at/5 + ab/20) + (bt-bb)**2*bb*h*(3*at/4 + ab/4) + \
                        (bt-bb)*bb**2*h*(at + ab/2) + bb**3*h*(at/2 + ab/2) )
    return x2+y2


def frustumIxx(rb_0, rt_0, h, diamFlag=False):
    """This function returns a frustum's mass-moment of inertia (divided by density) about the
    transverse x/y-axis passing through the center of mass with radii or diameter inputs.
    NOTE: This is for a SOLID frustum, not a shell

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    Ixx=Iyy : float (scalar/vector),  Moment of inertia about x/y-axis through center of mass (principle axes)
    """
    if diamFlag:
        # Convert diameters to radii
        rb, rt = 0.5 * rb_0, 0.5 * rt_0
    else:
        rb, rt = rb_0, rt_0
    # Integrate pi*r(z)^4/4 + pi*r(z)^2*(z-z_cg)^2 dz from z=0 to h
    A = 0.5 * frustumIzz(rb_0, rt_0, h)
    B = (
        np.pi
        * h**3
        / 80.0
        * (
            (rb**4 + 4.0 * rb**3 * rt + 10.0 * rb**2 * rt**2 + 4.0 * rb * rt**3 + rt**4)
            / (rb**2 + rb * rt + rt**2)
        )
    )
    return A + B


def RectangularFrustumIxx(ab, bb, at, bt, h):
    """This function returns a rectangular frustum's mass-moment of inertia (divided by density) about the
    central (axial) z-axis with side lengths and height
    NOTE: This is for a SOLID frustum, not a shell

    INPUTS:
    Parameters
    ----------
    ab : float (scalar/vector),  base side length a
    bb : float (scalar/vector),  base side length b
    at : float (scalar/vector),  top side length a
    bt : float (scalar/vector),  top side length b
    h  : float (scalar/vector),  height

    OUTPUTs:
    -------
    Izz : float (scalar/vector),  Moment of inertia about z-axis
    """

    # Use calculations from RAFT

    y2 = (1/12) * ( (bt-bb)**3*h*(at/5 + ab/20) + (bt-bb)**2*bb*h*(3*at/4 + ab/4) + \
                        (bt-bb)*bb**2*h*(at + ab/2) + bb**3*h*(at/2 + ab/2) )
    
    z2 = ( bt*at/5 + bb*at/20 + ab*bt/20 + bb*ab*(1/30) ) * h**3

    return y2 + z2

def RectangularFrustumIyy(ab, bb, at, bt, h):
    """This function returns a rectangular frustum's mass-moment of inertia (divided by density) about the
    central (axial) z-axis with side lengths and height
    NOTE: This is for a SOLID frustum, not a shell

    INPUTS:
    Parameters
    ----------
    ab : float (scalar/vector),  base side length a
    bb : float (scalar/vector),  base side length b
    at : float (scalar/vector),  top side length a
    bt : float (scalar/vector),  top side length b
    h  : float (scalar/vector),  height

    OUTPUTs:
    -------
    Izz : float (scalar/vector),  Moment of inertia about z-axis
    """

    # Use calculations from RAFT

    x2 = (1/12) * ( (at-ab)**3*h*(bt/5 + bb/20) + (at-ab)**2*ab*h*(3*bt/4 + bb/4) + \
                    (at-ab)*ab**2*h*(bt + bb/2) + ab**3*h*(bt/2 + bb/2) )
   
    z2 = ( bt*at/5 + bb*at/20 + ab*bt/20 + bb*ab*(1/30) ) * h**3

    return x2 + z2


def frustumShellVol(rb_0, rt_0, t, h, diamFlag=False):
    """This function returns a frustum shell's volume (for computing mass with density) with radii or diameter inputs.
    NOTE: This is for a frustum SHELL, not a solid

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    t  : float (scalar/vector),  thickness
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    cg : float (scalar/vector),  center of mass/gravity (ventroid)
    """
    if diamFlag:
        # Convert diameters to radii
        rb, rt = 0.5 * rb_0, 0.5 * rt_0
    else:
        rb, rt = rb_0, rt_0
    # Integrate 2*pi*r*dr*dz from r=ri(z) to ro(z), z=0 to h
    rb_o = rb
    rb_i = rb - t
    rt_o = rt
    rt_i = rt - t
    # ( (np.pi*h/3.0) * ( (rb_o**2 + rb_o*rt_o + rt_o**2) - (rb_i**2 + rb_i*rt_i + rt_i**2) ) )
    return frustumVol(rb_o, rt_o, h) - frustumVol(rb_i, rt_i, h)

def RectangularFrustumShellVol(ab, bb, at, bt, t, h):
    """This function returns a rectangular frustum shell's volume (for computing mass with density) with side lengths.
    NOTE: This is for a frustum SHELL, not a solid

    INPUTS:
    Parameters
    ----------
    ab : float (scalar/vector),  base side length a
    bb : float (scalar/vector),  base side length b
    at : float (scalar/vector),  top side length a
    bt : float (scalar/vector),  top side length b
    h  : float (scalar/vector),  height
    t  : float (scalar/vector),  thickness

    OUTPUTs:
    -------
    vol : float (scalar/vector),  shell volume
    """

    ab_o = ab
    bb_o = bb
    at_o = at
    bt_o = bt
    ab_i = ab_o - 2*t
    bb_i = bb_o - 2*t
    at_i = at_o - 2*t
    bt_i = bt_o - 2*t

    if isinstance(ab_i, float):
        assert ab_i > 0.0, "Thickness is larger than inner side length ab"
        assert bb_i > 0.0, "Thickness is larger than inner side length bb"
        assert at_i > 0.0, "Thickness is larger than inner side length at"
        assert bt_i > 0.0, "Thickness is larger than inner side length bt"
    else:
        assert (ab_i > 0.0).all(), "Thickness is larger than inner side length ab"
        assert (bb_i > 0.0).all(), "Thickness is larger than inner side length bb"
        assert (at_i > 0.0).all(), "Thickness is larger than inner side length at"
        assert (bt_i > 0.0).all(), "Thickness is larger than inner side length bt"

    return RectangularFrustumVol(ab_o, bb_o, at_o, bt_o, h) - RectangularFrustumVol(ab_i, bb_i, at_i, bt_i, h)


def frustumShellCG(rb_0, rt_0, t, h, diamFlag=False):
    """This function returns a frustum's center of mass/gravity (centroid) with radii or diameter inputs.
    NOTE: This is for a frustum SHELL, not a solid

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    t  : float (scalar/vector),  thickness
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    cg : float (scalar/vector),  center of mass/gravity (ventroid)
    """
    if diamFlag:
        # Convert diameters to radii
        rb, rt = 0.5 * rb_0, 0.5 * rt_0
    else:
        rb, rt = rb_0, rt_0
    # Integrate 2*pi*r*z*dr*dz/V from r=ri(z) to ro(z), z=0 to h
    rb_o = rb
    rb_i = rb - t
    rt_o = rt
    rt_i = rt - t
    A = (rb_o**2 + 2.0 * rb_o * rt_o + 3.0 * rt_o**2) - (rb_i**2 + 2.0 * rb_i * rt_i + 3.0 * rt_i**2)
    B = (rb_o**2 + rb_o * rt_o + rt_o**2) - (rb_i**2 + rb_i * rt_i + rt_i**2)
    return h * A / 4.0 / B

def RectangularFrustumShellCG(ab, bb, at, bt, t, h):
    """This function returns a rectangular frustum's center of mass/gravity (centroid) with side lengths and height
    NOTE: This is for a frustum SHELL, not a solid

    INPUTS:
    Parameters
    ----------
    ab : float (scalar/vector),  base side length a
    bb : float (scalar/vector),  base side length b
    at : float (scalar/vector),  top side length a
    bt : float (scalar/vector),  top side length b
    h  : float (scalar/vector),  height
    t  : float (scalar/vector),  thickness

    OUTPUTs:
    -------
    cg : float (scalar/vector),  center of mass/gravity (ventroid)
    """
    z_integral = 2.0/3.0*(at+bt+0.5*ab+0.5*bb)*h**2*t - 2.0*t**2*h**2
    return z_integral/RectangularFrustumShellVol(ab, bb, at, bt, t, h)


def frustumShellIzz(rb_0, rt_0, t, h, diamFlag=False):
    """This function returns a frustum's mass-moment of inertia (divided by density) about the
    central (axial) z-axis with radii or diameter inputs.
    NOTE: This is for a frustum SHELL, not a solid

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    t  : float (scalar/vector),  thickness
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    Izz : float (scalar/vector),  Moment of inertia about z-axis
    """
    if diamFlag:
        # Convert diameters to radii
        rb, rt = 0.5 * rb_0, 0.5 * rt_0
    else:
        rb, rt = rb_0, rt_0
    # Integrate 2*pi*r*dr*dz from r=ri(z) to ro(z), z=0 to h
    rb_o = rb
    rb_i = rb - t
    rt_o = rt
    rt_i = rt - t
    return frustumIzz(rb_o, rt_o, h) - frustumIzz(rb_i, rt_i, h)

def RectangularFrustumShellIzz(ab, bb, at, bt, t, h):
    """This function returns a rectangular frustum's mass-moment of inertia (divided by density) about the
    central (axial) z-axis with side lengths
    NOTE: This is for a frustum SHELL, not a solid

    INPUTS:
    Parameters
    ----------
    ab : float (scalar/vector),  base side length a
    bb : float (scalar/vector),  base side length b
    at : float (scalar/vector),  top side length a
    bt : float (scalar/vector),  top side length b
    h  : float (scalar/vector),  height
    t  : float (scalar/vector),  thickness

    OUTPUTs:
    -------
    Izz : float (scalar/vector),  Moment of inertia about z-axis
    """
    ab_o = ab
    bb_o = bb
    at_o = at
    bt_o = bt
    ab_i = ab_o - 2*t
    bb_i = bb_o - 2*t
    at_i = at_o - 2*t
    bt_i = bt_o - 2*t

    return RectangularFrustumIzz(ab_o, bb_o, at_o, bt_o, h) - RectangularFrustumIzz(ab_i, bb_i, at_i, bt_i, h)


def frustumShellIxx(rb_0, rt_0, t, h, diamFlag=False):
    """This function returns a frustum's mass-moment of inertia (divided by density) about the
    transverse x/y-axis passing through the center of mass with radii or diameter inputs.
    NOTE: This is for a frustum SHELL, not a solid

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    t  : float (scalar/vector),  thickness
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    Ixx=Iyy : float (scalar/vector),  Moment of inertia about x/y-axis through center of mass (principle axes)
    """
    if diamFlag:
        # Convert diameters to radii
        rb, rt = 0.5 * rb_0, 0.5 * rt_0
    else:
        rb, rt = rb_0, rt_0
    # Integrate 2*pi*r*dr*dz from r=ri(z) to ro(z), z=0 to h
    rb_o = rb
    rb_i = rb - t
    rt_o = rt
    rt_i = rt - t
    return frustumIxx(rb_o, rt_o, h) - frustumIxx(rb_i, rt_i, h)

def RectangularFrustumShellIxx(ab, bb, at, bt, t, h):
    """This function returns a rectangular frustum's mass-moment of inertia (divided by density) about the
    central (axial) z-axis with side lengths and height
    NOTE: This is for a frustum SHELL, not a solid

    INPUTS:
    Parameters
    ----------
    ab : float (scalar/vector),  base side length a
    bb : float (scalar/vector),  base side length b
    at : float (scalar/vector),  top side length a
    bt : float (scalar/vector),  top side length b
    h  : float (scalar/vector),  height
    t  : float (scalar/vector),  thickness

    OUTPUTs:
    -------
    Ixx : float (scalar/vector),  Moment of inertia about z-axis
    """
    ab_o = ab
    bb_o = bb
    at_o = at
    bt_o = bt
    ab_i = ab_o - 2*t
    bb_i = bb_o - 2*t
    at_i = at_o - 2*t
    bt_i = bt_o - 2*t

    return RectangularFrustumIxx(ab_o, bb_o, at_o, bt_o, h) - RectangularFrustumIxx(ab_i, bb_i, at_i, bt_i, h)

def RectangularFrustumShellIyy(ab, bb, at, bt, t, h):
    """This function returns a rectangular frustum's mass-moment of inertia (divided by density) about the
    central (axial) z-axis with side lengths and height
    NOTE: This is for a frustum SHELL, not a solid

    INPUTS:
    Parameters
    ----------
    ab : float (scalar/vector),  base side length a
    bb : float (scalar/vector),  base side length b
    at : float (scalar/vector),  top side length a
    bt : float (scalar/vector),  top side length b
    h  : float (scalar/vector),  height
    t  : float (scalar/vector),  thickness

    OUTPUTs:
    -------
    Izz : float (scalar/vector),  Moment of inertia about z-axis
    """
    ab_o = ab
    bb_o = bb
    at_o = at
    bt_o = bt
    ab_i = ab_o - 2*t
    bb_i = bb_o - 2*t
    at_i = at_o - 2*t
    bt_i = bt_o - 2*t

    return RectangularFrustumIyy(ab_o, bb_o, at_o, bt_o, h) - RectangularFrustumIyy(ab_i, bb_i, at_i, bt_i, h)


if __name__ == "__main__":
    Db = 6.5
    Dt = 4.0
    H = 120.0

    print("From commonse.Frustum: Sample Volume and CM of FRUSTUM=" + 4 * "{:8.4f}, ").format(
        *frustum(Db, Dt, H)[0].flatten()
    )


def main():
    pass



from unittest import TestCase
import os

import pandas as pd

from wisdem.landbosse.model import Cable, Array, ArraySystem
from wisdem.test.test_landbosse.test_WeatherDelay import generate_a_year
from wisdem.test.test_landbosse.test_filename_functions import landbosse_test_input_dir

import pytest

pd.set_option('display.width', 6000)
pd.set_option('display.max_columns', 20)

@pytest.mark.skip(reason="this does not pass")
class TestCollectionCost(TestCase):
    """
#
#     Tests CollectionCost module
#
#     """

    def setUp(self):
        self.input_dict = dict()
        self.output_dict = dict()
        self.cable_input_dict = dict()
        self.addl_specs = dict()

        rsmeans_csv = os.path.join(landbosse_test_input_dir(), 'rsmeans_proprietary.csv')
        self.input_dict['rsmeans'] = pd.read_csv(rsmeans_csv)
        crew_cost_csv = os.path.join(landbosse_test_input_dir(), 'crew_price_proprietary.csv')
        self.input_dict['crew_cost'] = pd.read_csv(crew_cost_csv)
        crews_csv = os.path.join(landbosse_test_input_dir(), 'crews.csv')
        self.input_dict['crew'] = pd.read_csv(crews_csv)
        self.input_dict['construct_duration'] = 9  # months
        self.input_dict['time_construct'] = 'normal'
        self.input_dict['hour_day'] = {'long': 24, 'normal': 10}

        #============================================================
        #All below inputs are inputs for ArraySystem class:
        self.input_dict['cable_specs'] = dict()


        cable_specs_path = os.path.join(landbosse_test_input_dir(), 'cable_specs.csv')

        self.input_dict['cable_specs_pd'] = pd.read_csv(cable_specs_path) #Read in cable specs into a dataframe.
        # self.input_dict['cable_specs'] = self.input_dict['cable_specs_pd'].T.to_dict()

        self.input_dict['user_defined_distance_to_grid_connection'] = 0  # 0 = No ; 1 = Yes
        self.input_dict['distance_to_grid_connection_km'] = 2
        self.project_name = 'project_1'

        #Inputs for ArraySystem:

        self.input_dict['num_turbines'] = 1
        self.input_dict['plant_capacity_MW'] = 15
        self.input_dict['row_spacing_rotor_diameters'] = 5
        self.input_dict['turbine_rating_MW'] = 0.02
        self.input_dict['upstream_turb'] = 0
        self.input_dict['turb_sequence'] = 1
        self.input_dict['depth'] = 45
        self.input_dict['rsmeans_per_diem'] = 144
        

        self.input_dict['turbine_spacing_rotor_diameters'] = 5
        self.input_dict['rotor_diameter_m'] = 154
        self.input_dict['line_frequency_hz'] = 60
        self.input_dict['overtime_multiplier'] = 1.4  # multiplier for labor overtime rates due to working 60 hr/wk rather than 40 hr/wk


        # Weather window:
        self.input_dict['num_delays'] = 7
        self.input_dict['avg_hours_per_delay'] = 20
        self.input_dict['std_dev_hours_per_delay'] = 5
        self.input_dict['delay_speed_m_per_s'] = 9
        self.input_dict['seed'] = 101
        self.input_dict['weather_window'] = generate_a_year(num_delays=self.input_dict['num_delays'],
                                                            avg_hours_per_delay=self.input_dict['avg_hours_per_delay'],
                                                            std_dev_hours_per_delay=self.input_dict[
                                                                'std_dev_hours_per_delay'],
                                                            delay_speed_m_per_s=self.input_dict['delay_speed_m_per_s'],
                                                            seed=self.input_dict['seed'])

        self.input_dict['start_delay_hours'] = 100
        self.input_dict['mission_time_hours'] = 50
        self.input_dict['critical_wind_speed_m_per_s'] = 8.0
        self.input_dict['wind_height_of_interest_m'] = 100
        self.input_dict['wind_shear_exponent'] = 0.25
        self.input_dict['season_construct'] = ['spring', 'summer']
        self.input_dict['time_construct'] = 'normal'
        self.input_dict['operational_hrs_per_day'] = 10
        self.input_dict['duration_construction'] = 9  # months
        self.input_dict['operational_hrs_per_day'] = 10
        self.input_dict['critical_speed_non_erection_wind_delays_m_per_s'] = 15
        self.input_dict['critical_height_non_erection_wind_delays_m'] = 10
        self.input_dict['rsmeans_per_diem'] = 99

        #============================================================



        # Test inputs for Cable class: (Taken from cable specs dict)
        self.cable_input_dict['Current Capacity (A)'] = 610  # Amps (at 1 m burial depth)
        self.cable_input_dict['Rated Voltage (V)'] = 33  # kV, line-to-line
        self.cable_input_dict['AC Resistance (Ohms/km)'] = 0.062  # Ohms/km (at 90 deg C at 60 Hz)
        self.cable_input_dict['Inductance (mH/km)'] = 0.381  # mH/km
        self.cable_input_dict['Capacitance (nF/km)'] = 224  # nF/km
        self.cable_input_dict['Cost (USD/LF)'] = 250000  # $US/km

        # Additional inputs for Array(Cable) class

        self.addl_specs['turbine_rating_MW'] = 7
        self.addl_specs['upstream_turb'] = 0
        self.addl_specs['turb_sequence'] = 0
        self.addl_specs['turbine_spacing_rotor_diameters'] = 5
        self.addl_specs['rotor_diameter_m'] = 154
        self.addl_specs['line_frequency_hz'] = 60


    def test_ArraySystemModule(self):
        """
        Black box test to check whether module is ran successfully or not
        """
        run_ArraySystem = ArraySystem(input_dict=self.input_dict, output_dict=self.output_dict, project_name=self.project_name)
        trial_run = run_ArraySystem.run_module()


        if trial_run[0] == 0 :
            print('\n\n================== MODULE EXECUTION SUCCESS =========================\n')
            print(' ArraySystem module ran successfully. See the list of inputs'
                  '\n and outputs below used by the module in its calculations:')
            print( '\n=====================================================================\n')

        elif trial_run[0] == 1 :
            print('\n\n====================================== MODULE EXECUTION FAILURE ======================================\n')
            print(' > ArraySystem module failed to run successfully. Error detected: ',trial_run[1] ,
                  '\n > Scroll below to see detailed information about error encountered.'
                  '\n > See the list of inputs below used by the module in its calculations:')
            print('\n========================================================================================================\n')


        print('\nGiven below is the set of inputs fed into ArraySystem module:\n')
        for key, value in self.input_dict.items():
            print(key, ':', value)


        if trial_run[0] == 0:  #Only print outputs if module ran successfully.
            print('\nGiven below is the set of outputs calculated by the ArraySystem module:\n')
            for k, v in self.output_dict.items():
                if isinstance(v, dict) :
                    for a, b in v.items():
                        if isinstance(b, pd.DataFrame):
                            return
                        else:
                            print('\nGiven below are attributes of cable ', a, ':\n')
                            for c, d in b.__dict__.items():
                                print(c, ':', d)
                            print('\n<----End of cable attributes---->\n')
                else:
                    if isinstance(v, pd.DataFrame):
                        print('\nNow printing DataFrame ->', k, ':\n', v)
                    else:
                        print(k, ':', v)


    def test_CableModule(self):

        """
        Black box test to check whether module is ran successfully or not
        """
        Cable(self.cable_input_dict, self.addl_specs)


    def test_ArrayModule(self):
        Array(self.cable_input_dict, self.addl_specs)


    def test_ArraySystem(self):
        run_collection = ArraySystem(self.input_dict, self.output_dict, project_name=self.project_name)
        run_collection.run_module()

    def test_calculate_weather_delay(self):
        """
        Tests calculate_weather_delay()

        """

        weatherDelay = ArraySystem(input_dict=self.input_dict, output_dict=self.output_dict, project_name=self.project_name)
        weatherDelay.run_module()





import os
import pickle
import openmdao
import numpy as np
import pandas as pd
import scipy.io as sio
from openmdao.utils.mpi import MPI

def get_variable_list(prob, rank_0=False):

    # Get all OpenMDAO inputs and outputs into a dictionary
    input_dict = prob.model.list_inputs(prom_name=True, units=True, desc=True, out_stream=None) # is_indep_var=True
    # If MPI, share input dictionary from rank 0 to all other ranks, which would otherwise be empty
    if MPI and rank_0 == False:
        input_dict = MPI.COMM_WORLD.bcast(input_dict, root=0) 
    for k in range(len(input_dict)):
        input_dict[k][1]["type"] = "input"

    inter_dict = prob.model.list_inputs(prom_name=True, units=True, desc=True, out_stream=None) # is_indep_var=False
    # If MPI, share intermediate dictionary from rank 0 to all other ranks, which would otherwise be empty
    if MPI and rank_0 == False:
        inter_dict = MPI.COMM_WORLD.bcast(inter_dict, root=0)
    for k in range(len(inter_dict)):
        inter_dict[k][1]["type"] = "intermediate"

    #var_dict = prob.model.list_inputs(prom_name=True, units=True, desc=True, out_stream=None)
    #for k in range(len(var_dict)):
    #    var_dict[k][1]["type"] = "output"

    out_dict = prob.model.list_outputs(prom_name=True, units=True, desc=True, out_stream=None)
    # If MPI, share output dictionary from rank 0 to all other ranks, which would otherwise be empty
    if MPI and rank_0 == False:
        out_dict = MPI.COMM_WORLD.bcast(out_dict, root=0)
    for k in range(len(out_dict)):
        out_dict[k][1]["type"] = "output"
        
    var_dict = input_dict.copy()
    var_dict.extend(inter_dict)
    var_dict.extend(out_dict)
    return input_dict, out_dict, var_dict


def variable_dict2df(var_dict):
    data = {}
    data["variables"] = []
    data["type"] = []
    data["units"] = []
    data["values"] = []
    data["description"] = []
    for k in range(len(var_dict)):
        unit_str = var_dict[k][1]["units"]
        if unit_str is None:
            unit_str = ""

        iname = var_dict[k][1]["prom_name"]
        itype = var_dict[k][1]["type"]
        if iname in data["variables"]:
            iprev = data["variables"].index( iname )
            if itype == "output":
                data["type"][iprev] = itype
            continue

        data["variables"].append(iname)
        data["type"].append(itype)
        data["units"].append(unit_str)
        data["values"].append(var_dict[k][1]["val"])
        data["description"].append(var_dict[k][1]["desc"])

    # Simplify stored value data type
    for k in range(len(data["values"])):
        if type(data["values"][k]) == type(np.array([])):
            if data["values"][k].size == 1:
                data["values"][k] = data["values"][k][0]
            else:
                data["values"][k] = data["values"][k].tolist()
        elif isinstance(data["values"][k], list) and len(data["values"][k]) == 1:
            data["values"][k] = data["values"][k][0]
                
    return pd.DataFrame(data)


def save_data(fname, prob, npz_file=True, mat_file=False, xls_file=True):
    # Get the variables
    _, _, var_dict = get_variable_list(prob, rank_0 = True)
    df = variable_dict2df(var_dict)
    
    # Remove file extension
    froot = os.path.splitext(fname)[0]

    # Reduce to variables we can save for matlab or python
    array_dict = {}
    for k in range(len(var_dict)):
        unit_str = var_dict[k][1]["units"]
        if unit_str is None or unit_str in ["Unavailable","n/a"]:
            unit_str = ""
        elif len(unit_str) > 0:
            unit_str = "_" + unit_str

        iname = var_dict[k][1]["prom_name"] + unit_str
        value = var_dict[k][1]["val"]

        if iname in array_dict:
            continue

        if type(value) in [type(np.array([])), type(0.0), type(0), np.float64, np.int64]:
            array_dict[iname] = value
        elif type(value) == type(True):
            array_dict[iname] = np.bool_(value)
        elif type(value) == type(""):
            array_dict[iname] = np.str_(value)
        elif type(value) == type([]):
            temp_val = np.empty(len(value), dtype=object)
            temp_val[:] = value[:]
            array_dict[iname] = temp_val
        # else:
        #    print(var_dict[k])

    # Pickle the full archive so that we can load it back in if we need
    df.to_pickle(froot + ".pkl")
        
    # Save to numpy compatible
    if npz_file:
        kwargs = {key: array_dict[key] for key in array_dict.keys()}
        np.savez_compressed(froot + ".npz", **kwargs)

    # Save to matlab compatible
    if mat_file:
        sio.savemat(froot + ".mat", array_dict, long_field_names=True)

    if xls_file:
        df.to_excel(froot + ".xlsx", index=False)
        df.to_csv(froot + ".csv", index=False)


    
        
def transfer_data(prob_from, prob_to, prefix_append=None, prefix_remove=None):

    if prefix_append is None:
        prefix_append = ''
        
    if prefix_remove is None:
        prefix_remove = ''

    # Get FROM data as DF
    if isinstance(prob_from, pd.DataFrame):
        var_df_from = prob_from
    elif isinstance(prob_from, openmdao.core.problem.Problem):
        _, _, var_dict_from = get_variable_list(prob_from)
        var_df_from = variable_dict2df(var_dict_from)
    
    # Get TO data as DF
    if isinstance(prob_to, pd.DataFrame):
        var_df_to = prob_to
    else:
        _, _, var_dict_to = get_variable_list(prob_to)
        var_df_to = variable_dict2df(var_dict_to)
    valid_vars_to = list(var_df_to["variables"])

    # Set the variables
    for k in range(len(var_df_from)):
        # Guess at the proper variable name
        invar_name = var_df_from["variables"].iloc[k]
        invar_units = var_df_from["units"].iloc[k]
        local_name = prefix_append + invar_name.replace(prefix_remove,"")
        local_name2 = local_name.replace(f"_{invar_units}", "")

        # Restore the type of the input data
        ival = var_df_from["values"].iloc[k]
        if isinstance(ival, (np.ndarray, float, int, list, dict)):
            pass
        elif isinstance(ival, str) and (ival.startswith('[') or ival.startswith('{')):
            ival = eval(ival)
        else:
            try:
                ival = float(ival)
            except Exception:
                pass

        # Store the value
        if local_name in valid_vars_to:
            prob_to[local_name] = ival
        elif local_name2 in valid_vars_to:
            prob_to[local_name2] = ival

    return prob_to


def load_data(fname, prob, prefix_append=None, prefix_remove=None):

    # Extracting from npz
    def npz2df(obj):
        data = {}
        data["variables"] = []
        data["units"] = []
        data["values"] = []
        for k in range(len(wt_obj.files)):
            iname = obj.files[k]
            iunit = iname.split('_')[-1]
            ival = obj[iname]
            data["variables"].append(iname)
            data["units"].append(iunit)
            data["values"].append(ival)
            
        mydf = pd.DataFrame(data)
        return mydf

    # Load in the difference filetypes
    if isinstance(fname, str) and fname.endswith(".xlsx"):
        mydf = pd.read_excel(fname)
        
    elif isinstance(fname, str) and fname.endswith(".csv"):
        mydf = pd.read_csv(fname)
        
    elif isinstance(fname, str) and fname.endswith(".npz"):
        wt_obj = np.load(fname, allow_pickle=True)
        mydf = npz2df(wt_obj)
        
    elif isinstance(fname, str) and fname.endswith(".pkl"):
        mydf = pd.read_pickle(fname)

    elif isinstance(fname, np.lib.npyio.NpzFile):
        mydf = npz2df(fname)

    elif isinstance(fname, pd.DataFrame) or isinstance(fname, openmdao.core.problem.Problem):
        mydf = fname

    else:
        raise Exception(f"Unknown file type, {fname}.  Expected xlsx or csv or npz or pkl")
    
    return transfer_data(mydf, prob, prefix_append=prefix_append, prefix_remove=prefix_remove)



import numpy as np
import openmdao.api as om
from scipy.optimize import minimize

# Application factor to include ring/housing/carrier weight
Kr = 0.4

# -----------------------------------


def V_planetary(U, B, K):
    sunU = 0.5 * U - 1.0
    V = (
        1.0 / U
        + 1.0 / U / B
        + 1.0 / B / sunU
        + sunU
        + sunU**2
        + K * (U - 1.0) ** 2 / B
        + K * (U - 1.0) ** 2 / B / sunU
    )
    return V


# -----------------------------------


def V_parallel(U):
    V = 1.0 + 1.0 / U + U + U**2
    return V


# -----------------------------------
def volumeEEP(x, n_planets, torque, Kr1=Kr, Kr2=Kr):
    # Safety factor?
    Kgamma = [1.1 if m < 5 else 1.35 for m in n_planets]

    # Individual stage torques
    Q_stage = torque / np.cumprod(x)

    # Volume
    V = (
        Q_stage[0] * Kgamma[0] * V_planetary(x[0], n_planets[0], Kr1)
        + Q_stage[1] * Kgamma[1] * V_planetary(x[1], n_planets[1], Kr2)
        + Q_stage[2] * V_parallel(x[2]) / np.prod(x)
    )
    return 2 * V


# -----------------------------------


def volumeEPP(x, n_planets, torque, Kr1=Kr):
    # Safety factor?
    Kgamma = [1.1 if m < 5 else 1.35 for m in n_planets]

    # Individual stage torques
    Q_stage = torque / np.cumprod(x)

    V = (
        Q_stage[0] * Kgamma[0] * V_planetary(x[0], n_planets[0], Kr1)
        + Q_stage[1] * V_parallel(x[1]) / np.prod(x[:2])
        + Q_stage[2] * V_parallel(x[2]) / np.prod(x)
    )
    return 2 * V


# -----------------------------------


class Gearbox(om.ExplicitComponent):
    """
    The gearbox design follows the general approach of the previous DriveSE implementation, however
    with code improvements, the results will likely be different than prior versions.  The gearbox is
    assumed to have 3 stages, with the user specifying a configuration code of either "EEP" or "EPP",
    with the "E" representing epicyclic (planetary) gear stages and "P" representing parallel gear stages.
    For the epicyclic stages, the user also has to specify the number of planets, so the EEP input would
    require something like [3, 3, 0] and EPP would require [3, 0, 0].  The user also specifies the overall
    target gear ratio, and then DrivetrainSE conducts a mass minimization of the three stage ratios that
    meet the target and minimize the overall mass.

    Parameters
    ----------
    gear_configuration : string
        3-letter string of Es or Ps to denote epicyclic or parallel gear configuration
    n_planets : numpy array[3]
        number of planets in each stage
    gear_ratio : float
        overall gearbox speedup ratio
    D_rotor : float, [m]
        rotor diameter
    Q_rotor : float, [N*m]
        rotor torque at rated power
    s_gearbox : float, [m]
        gearbox position along x-axis

    Returns
    -------
    stage_masses : numpy array[3], [kg]
        individual gearbox stage gearbox_masses
    gearbox_mass : float, [kg]
        overall component mass
    gearbox_cm : numpy array[3], [m]
        Gearbox center of mass [x,y,z] measure along shaft from bedplate
    gearbox_I : numpy array[3], [kg*m**2]
        Gearbox mass moments of inertia [Ixx, Iyy, Izz] around its center of mass
    L_gearbox : float, [m]
        length of gearbox
    H_gearbox : float, [m]
        height of gearbox
    D_gearbox : float, [m]
        diameter of gearbox

    """

    def initialize(self):
        self.options.declare("direct_drive", default=True)
        self.options.declare("use_gb_torque_density", default=True)

    def setup(self):
        self.add_discrete_input("gear_configuration", val="eep")
        # self.add_discrete_input('shaft_factor', val='normal')
        self.add_discrete_input("planet_numbers", val=[3, 3, 0])
        self.add_input("gear_ratio", val=1.0)
        self.add_input("rotor_diameter", val=0.0, units="m")
        self.add_input("rated_torque", val=0.0, units="N*m")
        self.add_input("machine_rating", val=0.0, units="kW")
        self.add_input("gearbox_mass_user", val=0.0, units="kg")
        self.add_input("gearbox_torque_density", val=0.0, units="N*m/kg")
        self.add_input("gearbox_radius_user", val=0.0, units="m")
        self.add_input("gearbox_length_user", val=0.0, units="m")

        self.add_output("stage_ratios", val=np.zeros(3))
        self.add_output("gearbox_mass", 0.0, units="kg")
        self.add_output("gearbox_I", np.zeros(3), units="kg*m**2")
        self.add_output("L_gearbox", 0.0, units="m")
        self.add_output("D_gearbox", 0.0, units="m")
        self.add_output("carrier_mass", 0.0, units="kg")
        self.add_output("carrier_I", np.zeros(3), units="kg*m**2")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        if self.options["direct_drive"]:
            outputs["stage_ratios"] = np.zeros(3)
            outputs["gearbox_mass"] = outputs["D_gearbox"] = outputs["L_gearbox"] = np.zeros(1)
            outputs["gearbox_I"] = np.zeros(3)
            return

        # Unpack inputs
        config = discrete_inputs["gear_configuration"]
        # shaft_factor = discrete_inputs['shaft_factor']
        n_planets = np.maximum(1.0, np.array(discrete_inputs["planet_numbers"]))
        gear_ratio = float(inputs["gear_ratio"][0])
        torque = float(inputs["rated_torque"][0])
        rating = float(inputs["machine_rating"][0])
        n_stage = 3
        
        # Other gearbox elements that are just estimates: shrink disc and carrier
        m_shrink_disc = rating / 3.0
        m_carrier = 8e3
        outputs["carrier_mass"] = m_shrink_disc + m_carrier
        
        # calculate mass properties
        D_rotor = float(inputs["rotor_diameter"][0])
        
        R_gearbox = float(inputs["gearbox_radius_user"][0])
        if R_gearbox == 0.0:
            R_gearbox = 0.006 * D_rotor # assumed to be 0.6% of wind turbine rotor diameter, regression from Jan 30th, 2024
                
        L_gearbox = float(inputs["gearbox_length_user"][0])
        if L_gearbox == 0.0:
            L_gearbox = 0.015 * D_rotor # assumed to be 1.5% of wind turbine rotor diameter, regression from Jan 30th, 2024

        # Moment of inertia (without mass yet)
        I = np.zeros(3)
        I[0] = 0.5 * R_gearbox**2
        I[1:] = (1.0 / 12.0) * (3 * R_gearbox**2 + L_gearbox**2)

        # Store some outputs
        outputs["D_gearbox"] = 2 * R_gearbox
        outputs["L_gearbox"] = L_gearbox
        outputs["carrier_I"] = outputs["carrier_mass"] * I[0] * np.array([1.0, 0.5, 0.5])  # Solid disk

        # Now determine gearbox mass
        m_gearbox = float(inputs["gearbox_mass_user"][0])
        
        if m_gearbox == 0.0 and self.options["use_gb_torque_density"]:
            # NOTE THIS IS DEFAULT BECAUSE WE TRUST IT MORE AND IT IS MUCH QUICKER
            m_gearbox = torque / float(inputs["gearbox_torque_density"][0])

        if m_gearbox == 0.0:

            # Known configuration checks
            if config.lower() not in ["eep", "eep_2", "eep_3", "epp"]:
                raise ValueError("Invalid value for gearbox_configuration.  Must be one of: eep, eep_2, eep_3, epp")

            # Optimize stage ratios
            # Use double sided constraints to hack inequality constraints as COBYLA seems to work better than SLSQP here
            def constr1(x, ratio):
                return np.prod(x) - ratio

            def constr2(x, ratio):
                return ratio - np.prod(x)

            x0 = gear_ratio ** (1.0 / n_stage) * np.ones(n_stage)
            bounds = [[2.01, 20.0], [2.01, 20.0], [2.01, 20.0]]
            const = [{}, {}]
            const[0]["type"] = "ineq"
            const[0]["fun"] = constr1
            const[0]["args"] = [gear_ratio]
            const[1]["type"] = "ineq"
            const[1]["fun"] = constr2
            const[1]["args"] = [gear_ratio]
            method = "cobyla"
            tol = 1e-3

            if config.lower() == "eep":
                bounds[2][0] = 1.0
                result = minimize(
                    lambda x: volumeEEP(x, n_planets, torque),
                    x0,
                    method=method,  # bounds=bounds,
                    tol=tol,
                    constraints=const,
                    options={"maxiter": 100, "disp": False},
                )
                ratio_stage = result.x

            elif config == "eep_3":
                # fixes last stage ratio at 3
                const[0]["args"] = const[1]["args"] = [gear_ratio / 3.0]
                bounds[2][0] = 1.0
                result = minimize(
                    lambda x: volumeEEP(np.r_[x, 3.0], n_planets, torque),
                    x0[:2],
                    method=method,  # bounds=bounds,
                    tol=tol,
                    constraints=const,
                    options={"maxiter": 100, "disp": False},
                )
                ratio_stage = np.r_[result.x, 3.0]

            elif config == "eep_2":
                # fixes final stage ratio at 2
                const[0]["args"] = const[1]["args"] = [gear_ratio / 2.0]
                bounds[2][0] = 1.0
                result = minimize(
                    lambda x: volumeEEP(np.r_[x, 2.0], n_planets, torque),
                    x0[:2],
                    method=method,  # bounds=bounds,
                    tol=tol,
                    constraints=const,
                    options={"maxiter": 100, "disp": False},
                )
                ratio_stage = np.r_[result.x, 2.0]

            elif config == "epp":
                bounds[1][0] = 1.0
                bounds[2][0] = 1.0
                result = minimize(
                    lambda x: volumeEPP(x, n_planets, torque),
                    x0,
                    method=method,  # bounds=bounds,
                    tol=tol,
                    constraints=const,
                    options={"maxiter": 100, "disp": False},
                )
                ratio_stage = result.x

            # Get final volume
            if config.lower().find("eep") >= 0:
                vol = volumeEEP(ratio_stage, n_planets, torque)
            else:
                vol = volumeEPP(ratio_stage, n_planets, torque)
            outputs["stage_ratios"] = ratio_stage

            # Cumulative Kfactor scaling based on values reported in Nejad's paper even though that was probably done with the buggy version of DriveSE
            K = np.mean([48.82 / 15216.504, 53.69 / 17401.453])

            # Shaft length factor
            Kshaft = 1.0  # if shaft_factor == 'normal' else 1.25

            # All factors into the mass
            m_gearbox = K * Kshaft * vol.sum()

            # Other gearbox elements that are just estimates: shrink disc and carrier
            m_gearbox += m_shrink_disc + m_carrier

        # Store outputs
        outputs["gearbox_mass"] = m_gearbox
        outputs["gearbox_I"] = I * m_gearbox



import os
import numpy as np

class PreCompWriter:
    def __init__(self, dir_out, materials, upper, lower, webs, profile, chord, twist, p_le):
        self.dir_out = dir_out

        self.materials = materials
        self.upper = upper
        self.lower = lower
        self.webs = webs
        self.profile = profile

        self.chord = chord
        self.twist = twist
        self.p_le = p_le

        if not os.path.exists(dir_out):
            os.makedirs(dir_out)

    def execute(self):
        flist_layup = self.writePreCompLayup()
        flist_profile = self.writePreCompProfile()
        self.writePreCompMaterials()
        self.writePreCompInput(flist_layup, flist_profile)

    def writePreCompMaterials(self):
        text = []
        text.append("\n")
        text.append("Mat_Id     E1           E2          G12       Nu12     Density      Mat_Name\n")
        text.append(" (-)      (Pa)         (Pa)        (Pa)       (-)      (Kg/m^3)       (-)\n")

        for i, mat in enumerate(self.materials):
            text.append("%d %e %e %e %f %e %s\n" % (i + 1, mat.E1, mat.E2, mat.G12, mat.nu12, mat.rho, mat.name))

        fout = os.path.join(self.dir_out, "materials.inp")
        f = open(fout, "w")
        for outLine in text:
            f.write(outLine)
        f.close()

    def writePreCompLayup(self):
        f_out = []

        def write_layup_sectors(cs, web):
            text = []
            for i, (n_plies, t, theta, mat_idx) in enumerate(zip(cs.n_plies, cs.t, cs.theta, cs.mat_idx)):
                if web:
                    text.extend(["\n", "web_num    no of laminae (N_weblams)    Name of stack:\n"])
                else:
                    text.extend(
                        [
                            "...........................................................................\n",
                            "Sect_num    no of laminae (N_laminas)          STACK:\n",
                        ]
                    )
                n_lamina = len(n_plies)
                text.append("%d %d\n" % (i + 1, n_lamina))
                text.extend(
                    [
                        "\n",
                        "lamina    num of  thickness   fibers_direction  composite_material ID\n",
                        "number    plies   of ply (m)       (deg)               (-)\n",
                    ]
                )
                if web:
                    text.append("wlam_num N_Plies   w_tply       Tht_Wlam            Wmat_Id\n")
                else:
                    text.append("lam_num  N_plies    Tply         Tht_lam            Mat_id\n")

                for j, (plies_j, t_j, theta_j, mat_idx_j) in enumerate(zip(n_plies, t, theta, mat_idx + 1)):
                    text.append("%d %d %e %.1f %d\n" % (j + 1, plies_j, t_j, theta_j, mat_idx_j))
            return text

        for idx, (lower_i, upper_i, webs_i) in enumerate(zip(self.lower, self.upper, self.webs)):
            text = []
            text.append("Composite laminae lay-up inside the blade section\n")
            text.append("\n")
            text.append("*************************** TOP SURFACE ****************************\n")
            # number of sectors
            n_sector = len(upper_i.loc) - 1
            text.append("%d                N_scts(1):  no of sectors on top surface\n" % n_sector)
            text.extend(["\n", "normalized chord location of  nodes defining airfoil sectors boundaries (xsec_node)\n"])
            locU = upper_i.loc
            text.append(" ".join(["%f" % i for i in locU]) + "\n")
            text.extend(write_layup_sectors(upper_i, False))

            text.extend(["\n", "\n", "*************************** BOTTOM SURFACE ****************************\n"])
            n_sector = len(lower_i.loc) - 1
            text.append("%d                N_scts(1):  no of sectors on top surface\n" % n_sector)
            text.extend(["\n", "normalized chord location of  nodes defining airfoil sectors boundaries (xsec_node)\n"])
            locU = lower_i.loc
            text.append(" ".join(["%f" % i for i in locU]) + "\n")
            text.extend(write_layup_sectors(lower_i, False))

            text.extend(
                [
                    "\n",
                    "\n",
                    "**********************************************************************\n",
                    "Laminae schedule for webs (input required only if webs exist at this section):\n",
                ]
            )
            ########## Webs ##########
            text.extend(write_layup_sectors(webs_i, True))

            fname = os.path.join(self.dir_out, "layup_%00d.inp" % idx)
            f = open(fname, "w")
            for outLine in text:
                f.write(outLine)
            f.close()
            f_out.append(fname)

        return f_out

    def writePreCompProfile(self):
        f_out = []
        for idx, profile_i in enumerate(self.profile):
            # idx = 0
            # profile_i = profile[idx]
            text = []

            text.append(
                "%d                      N_af_nodes :no of airfoil nodes, counted clockwise starting\n"
                % len(profile_i.x)
            )
            text.append("                      with leading edge (see users' manual, fig xx)\n")
            text.append("\n")
            text.append(" Xnode      Ynode   !! chord-normalized coordinated of the airfoil nodes\n")

            x_all = np.concatenate((profile_i.x, np.flip(profile_i.x, 0)))
            y_all = np.concatenate((profile_i.yu, np.flip(profile_i.yl, 0)))

            if max(y_all) > 1.0:
                print(idx)

            # import matplotlib.pyplot as plt
            # plt.figure()
            # plt.plot(x_all, y_all)
            # plt.savefig('test.png')

            for x, y in zip(x_all, y_all):
                text.append("%f %f\n" % (x, y))

            fname = os.path.join(self.dir_out, "shape_%d.inp" % idx)
            f = open(fname, "w")
            for outLine in text:
                f.write(outLine)
            f.close()
            f_out.append(fname)

        return f_out

    def writePreCompInput(self, flist_layup, flist_profile):
        for idx in range(0, len(flist_layup)):
            chord = self.chord[idx]
            twist = self.twist[idx]
            p_le = self.p_le[idx]
            webs_i = self.webs[idx]
            layup_i = os.path.split(os.path.abspath(flist_layup[idx]))[1]
            profile_i = os.path.split(os.path.abspath(flist_profile[idx]))[1]

            text = []
            text.append("*****************  main input file for PreComp *****************************\n")
            text.append("Sample Composite Blade Section Properties\n")
            text.append("\n")
            text.append("General information -----------------------------------------------\n")
            text.append("1                Bl_length   : blade length (m)\n")
            text.append("2                N_sections  : no of blade sections (-)\n")
            text.append(
                "%d                N_materials : no of materials listed in the materials table (material.inp)\n"
                % len(self.materials)
            )
            text.append("3                Out_format  : output file   (1: general format, 2: BModes-format, 3: both)\n")
            text.append("f                TabDelim     (true: tab-delimited table; false: space-delimited table)\n")
            text.append("\n")
            text.append("Blade-sections-specific data --------------------------------------\n")
            text.append("Sec span     l.e.     chord   aerodynamic   af_shape    int str layup\n")
            text.append("location   position   length    twist         file          file\n")
            text.append("Span_loc    Le_loc    Chord    Tw_aero   Af_shape_file  Int_str_file\n")
            text.append("  (-)        (-)       (m)    (degrees)       (-)           (-)\n")
            text.append("\n")
            text.append("%.2f %f %e %f %s %s\n" % (0.0, p_le, chord, twist, profile_i, layup_i))
            text.append("%.2f %f %e %f %s %s\n" % (1.0, p_le, chord, twist, profile_i, layup_i))
            text.append("\n")
            text.append("Webs (spars) data  --------------------------------------------------\n")
            text.append("\n")
            text.append(
                "%d                Nweb        : number of webs (-)  ! enter 0 if the blade has no webs\n"
                % len(webs_i.loc)
            )
            text.append(
                "1                Ib_sp_stn   : blade station number where inner-most end of webs is located (-)\n"
            )
            text.append(
                "2                Ob_sp_stn   : blade station number where outer-most end of webs is located (-)\n"
            )
            text.append("\n")
            text.append("Web_num   Inb_end_ch_loc   Oub_end_ch_loc (fraction of chord length)\n")
            for i, loc in enumerate(webs_i.loc):
                text.append("%d %f %f\n" % (i + 1, loc, loc))

            fname = os.path.join(self.dir_out, "input_%d.inp" % idx)
            f = open(fname, "w")
            for outLine in text:
                f.write(outLine)
            f.close()




#!/usr/bin/env python3
import os

from wisdem import run_wisdem

## File management
mydir = os.path.dirname(os.path.realpath(__file__))  # get path to this file
fname_wt_input = mydir + os.sep + "nrel5mw.yaml"
fname_modeling_options = mydir + os.sep + "modeling_options.yaml"
fname_analysis_options = mydir + os.sep + "analysis_options.yaml"

wt_opt, analysis_options, opt_options = run_wisdem(fname_wt_input, fname_modeling_options, fname_analysis_options)
# end

print("blade mass:", wt_opt["tcc.blade_mass"])
print("blade moments of inertia:", wt_opt["drivese.blades_I"])
print("BRFM:", wt_opt["drivese.pitch_system.BRFM"])
print("hub forces:", wt_opt["drivese.F_aero_hub"])
print("hub moments:", wt_opt["drivese.M_aero_hub"])



import os

from wisdem import run_wisdem
from openmdao.utils.mpi import MPI
from wisdem.postprocessing.compare_designs import run

mydir = os.path.dirname(os.path.realpath(__file__))  # get path to this file
fname_wt_input = os.path.join(os.path.dirname(mydir), "02_reference_turbines", "IEA-15-240-RWT.yaml")
fname_modeling_options = os.path.join(os.path.dirname(mydir), "02_reference_turbines", "modeling_options.yaml")
fname_analysis_options = mydir + os.sep + "analysis_options.yaml"

wt_opt, modeling_options, analysis_options = run_wisdem(fname_wt_input, fname_modeling_options, fname_analysis_options)

if MPI:
    rank = MPI.COMM_WORLD.Get_rank()
else:
    rank = 0

if rank == 0:
    print(
        "RUN COMPLETED. RESULTS ARE AVAILABLE HERE: "
        + os.path.join(mydir, analysis_options["general"]["folder_output"])
    )

run([wt_opt], ["optimized"], modeling_options, analysis_options)



"""
A test where we can modify different parts of the modeling_options to ensure
they work with a top-to-bottom WISDEM run.
"""

import os
import unittest

from wisdem.glue_code.runWISDEM import run_wisdem

test_dir = os.path.join(
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))),
    "examples",
    "09_floating"
)
fname_analysis_options = os.path.join(test_dir, "analysis_options.yaml")
fname_modeling_options = os.path.join(test_dir, "modeling_options.yaml")


class TestRegression(unittest.TestCase):
    def test15MW(self):
        ## IEA 15MW
        fname_wt_input = os.path.join(test_dir, "IEA-15-240-RWT_VolturnUS-S.yaml")
        wt_opt, modeling_options, opt_options = run_wisdem(
            fname_wt_input, fname_modeling_options, fname_analysis_options
        )

        self.assertAlmostEqual(wt_opt["rotorse.rp.AEP"][0] * 1.0e-6, 78.00735886752878, 1)
        self.assertAlmostEqual(wt_opt["rotorse.blade_mass"][0], 68375.20114828358, -1)
        self.assertAlmostEqual(wt_opt["financese.lcoe"][0] * 1.0e3, 85.55178788445798, 1)
        self.assertAlmostEqual(wt_opt["rotorse.rs.tip_pos.tip_deflection"][0], 26.179096107022136, 1)


if __name__ == "__main__":
    unittest.main()



from unittest import TestCase
import pandas as pd
import numpy as np
import pytest

from wisdem.landbosse.model import WeatherDelay


SEASON_WINTER = 'winter'
SEASON_SPRING = 'spring'
SEASON_SUMMER = 'summer'
SEASON_FALL = 'fall'

month_numbers_to_seasons = {
    1: SEASON_WINTER,
    2: SEASON_WINTER,
    3: SEASON_WINTER,
    4: SEASON_SPRING,
    5: SEASON_SPRING,
    6: SEASON_SPRING,
    7: SEASON_SUMMER,
    8: SEASON_SUMMER,
    9: SEASON_SUMMER,
    10: SEASON_FALL,
    11: SEASON_FALL,
    12: SEASON_FALL
}

days_in_months = {
    'jan': 31,
    'feb': 28,
    'mar': 31,
    'apr': 30,
    'may': 31,
    'jun': 30,
    'jul': 31,
    'aug': 31,
    'sep': 30,
    'oct': 31,
    'nov': 30,
    'dec': 31
}


def generate_a_year(num_delays,
                    avg_hours_per_delay,
                    std_dev_hours_per_delay,
                    delay_speed_m_per_s,
                    seed):
    """
    This function generates a year of data. It creates a dataframe with
    the columns: 'Year', 'Month', 'Day', 'Hour', 'Season', 'Time window'

    Parameters
    ----------
    num_delays : int
        Number of delays in the year

    avg_hours_per_delay : int
        Average duration in hours for each delay.

    std_dev_hours_per_delay : int
        The standard deviation of blackout durations

    delay_speed_m_per_s : float
        Wind speed that will cause a weather delay. This may be separate
        from the critical wind speed for the crane. The critical wind speed
        for the crane is calculated according to the wind shear power law.
        The delay speed is the reference wind speed.

    seed : int
        NumPy random number generator seed. All the data is generated by
        random number generators. Byt giving the same value for this seed
        you can reproduce the same dataframe repeatedly for testing.

    Returns
    -------
    list
        A list of dictionaries. Each dictionary has the following keys
        month, day, year
    """
    np.random.seed(seed)
    year = []
    for month, days_in_month in zip(range(1, 13), days_in_months.values()):
        for day in range(1, days_in_month + 1):
            for hour in range(24):
                row = dict()
                row['Month'] = month
                row['Day'] = day
                row['Hour'] = hour
                row['Season'] = month_numbers_to_seasons[month]
                row['Time window'] = 'normal' if 8 <= hour <= 18 else 'long'
                year.append(row)
    speed_m_per_s = np.zeros(len(year))
    durations = np.random.normal(loc=avg_hours_per_delay, scale=std_dev_hours_per_delay, size=num_delays)
    durations = np.ceil(durations).astype('int')
    positions = np.random.uniform(0, len(year), num_delays)
    positions = np.ceil(positions).astype('int')
    for duration, position in zip(durations, positions):
        delay = np.empty(duration)
        delay.fill(delay_speed_m_per_s)
        speed_m_per_s[position:position + duration] = delay
    df = pd.DataFrame(data=year)
    df['Speed m per s'] = speed_m_per_s
    return df


class TestWeatherDelay(TestCase):
    def setUp(self):
        # Use all default parameters shown above, including the default
        # random seed.
        self.num_delays = 7
        self.avg_hours_per_delay = 20
        self.std_dev_hours_per_delay = 5
        self.delay_speed_m_per_s = 9
        self.seed = 101
        self.weather_window = generate_a_year(num_delays=self.num_delays,
                                              avg_hours_per_delay=self.avg_hours_per_delay,
                                              std_dev_hours_per_delay=self.std_dev_hours_per_delay,
                                              delay_speed_m_per_s=self.delay_speed_m_per_s,
                                              seed=self.seed)

    def test_count_delays_over_all_hours(self):
        """
        Tests delay detection count for 24 hours per day, 365 days per year.
        """
        weather_delay_input_dict = dict()
        weather_delay_input_dict['weather_window'] = self.weather_window
        weather_delay_input_dict['season_construct'] = ['winter', 'spring', 'summer', 'fall']
        weather_delay_input_dict['time_construct'] = ['normal', 'long']
        weather_delay_input_dict['start_delay_hours'] = 0
        weather_delay_input_dict['mission_time_hours'] = 8760
        weather_delay_input_dict['critical_wind_speed_m_per_s'] = 6.0
        weather_delay_input_dict['wind_height_of_interest_m'] = 25
        weather_delay_input_dict['wind_shear_exponent'] = 0.25
        output_dict = dict()
        wd = WeatherDelay(input_dict=weather_delay_input_dict, output_dict=output_dict)
        wind_delays = output_dict['wind_delays']
        self.assertEqual(self.num_delays, len(wind_delays), 'WeatherDelay does not count proper number of delays.')

    def test_delay_durations_over_all_hours(self):
        """
        Tests delay duration detection for 24 hours per day 365 days per year
        """
        weather_delay_input_dict = dict()
        weather_delay_input_dict['weather_window'] = self.weather_window
        weather_delay_input_dict['season_construct'] = ['winter', 'spring', 'summer', 'fall']
        weather_delay_input_dict['time_construct'] = ['normal', 'long']
        weather_delay_input_dict['start_delay_hours'] = 0
        weather_delay_input_dict['mission_time_hours'] = 8760
        weather_delay_input_dict['critical_wind_speed_m_per_s'] = 6.0
        weather_delay_input_dict['wind_height_of_interest_m'] = 25
        weather_delay_input_dict['wind_shear_exponent'] = 0.25
        output_dict = dict()
        wd = WeatherDelay(input_dict=weather_delay_input_dict, output_dict=output_dict)
        expected = [16, 25, 19, 24, 34, 23, 24]
        actual = output_dict['wind_delays']
        self.assertEqual(expected, actual, 'WeatherDelay does not match delay durations.')

    def test_keys_present(self):
        """
        This test deliberately passes a bad input dictionary to ensure that an
        exception is raised when the bad input dictionary does not pass
        validation.
        """
        bad_input_dict = dict()
        bad_input_dict['weather_window'] = self.weather_window
        bad_input_dict['season_construct'] = ['winter', 'spring', 'summer', 'fall']
        output_dict = dict()
        self.assertRaises(ValueError, WeatherDelay, bad_input_dict, output_dict)



#!/usr/bin/env python
import numpy as np
import os.path as osp

from wisdem import run_wisdem
from wisdem.commonse import fileIO
from get_docstrings import get_all_docstrings

# Get all docstrings from WISDEM files
parsed_dict = get_all_docstrings()

# Run comprehensive cases that should grab almost all inputs and outputs (might be missing some generators and BOS options)
examp_dir = osp.join(osp.dirname(osp.dirname(osp.dirname(osp.realpath(__file__)))), "examples")
mydir = osp.join(examp_dir, "02_reference_turbines")
iea15_mono_geom = osp.join(mydir, "IEA-15-240-RWT.yaml")
iea3p4_geom = osp.join(mydir, "IEA-3p4-130-RWT.yaml")
rwt_modeling = osp.join(mydir, "modeling_options.yaml")
rwt_analysis = osp.join(mydir, "analysis_options.yaml")

mydir = osp.join(examp_dir, "09_floating")
iea15_float_geom = osp.join(mydir, "IEA-15-240-RWT_VolturnUS-S.yaml")
rwt_float_modeling = osp.join(mydir, "modeling_options.yaml")
rwt_float_analysis = osp.join(mydir, "analysis_options.yaml")

mydir = osp.join(examp_dir, "17_jacket")
nrel_jack_geom = osp.join(mydir, "nrel5mw_jacket.yaml")
rwt_jack_modeling = osp.join(mydir, "modeling_options_jacket.yaml")
rwt_jack_analysis = osp.join(mydir, "analysis_options_jacket.yaml")

iea15_mono_prob, model_dict, anal_dict              = run_wisdem(iea15_mono_geom, rwt_modeling, rwt_analysis)
iea3p4_prob, _, _                                   = run_wisdem(iea3p4_geom    , rwt_modeling, rwt_analysis)
iea15_float_prob, model_float_dict, anal_float_dict = run_wisdem(iea15_float_geom, rwt_float_modeling, rwt_float_analysis)
nrel_jack_prob, model_jack_dict, anal_jack_dict     = run_wisdem(iea15_float_geom, rwt_float_modeling, rwt_float_analysis)

# Extract inputs and outputs from the models
all_inputs = []
all_outputs = []
for p in [iea15_mono_prob, iea3p4_prob, iea15_float_prob, nrel_jack_prob]:
    input_k, output_k, _ = fileIO.get_variable_list(p)
    if len(all_inputs) == 0:
        all_inputs = input_k
        all_outputs = output_k
    else:
        all_inputs.extend(input_k)
        all_outputs.extend(output_k)

# Use Pandas for some data cleansing and writing to csv
def write_guide(in_dict, fname):
    mydf = fileIO.variable_dict2df(in_dict)
    mydf.rename(columns={"variables":"Variable",
                         "units":"Units",
                         "description":"Description"}, inplace=True)
    mydf = mydf[["Variable", "Units", "Description"]]
    mydf.set_index("Variable", inplace=True)
    mydf = mydf[~mydf.index.duplicated(keep='first')]
    mydf.reset_index(inplace=True)

    # Fold in docstrings
    mynames = mydf["Variable"].to_list()
    mydesc  = mydf["Description"].to_list()
    for k in range(len(mynames)):
        ivar = mynames[k]
        if ivar in parsed_dict:
            idesc = parsed_dict[ivar]
            mydesc[k] += "" if idesc is None else idesc
    mydf["Description"] = mydesc
    mydf['Units'] = mydf['Units'].replace(np.nan, '-')
    mydf['Description'] = mydf['Description'].replace(np.nan,'None')

    # Write everything out
    mydf.to_csv(fname, index=False)
    mydf.to_json(fname.replace('csv','json'))#, index=False)

    return mydf

inputs_df = write_guide(all_inputs, "input_variable_guide.csv")
outputs_df = write_guide(all_outputs, "output_variable_guide.csv")
    



import os
import unittest

import numpy as np
import openmdao.api as om
import numpy.testing as npt

import wisdem.rotorse.rotor_power as rp

# Load in airfoil and blade shape inputs for NREL 5MW
ARCHIVE = os.path.dirname(os.path.abspath(__file__)) + os.path.sep + "regulation.npz"
NPZFILE = np.load(ARCHIVE)


def fillprob(prob, n_pc, n_span):
    prob.setup()
    for k in NPZFILE.files:
        prob[k] = NPZFILE[k]

    prob.set_val("v_min", 4.0, units="m/s")
    prob.set_val("v_max", 25.0, units="m/s")
    prob.set_val("rated_power", 5e6, units="W")
    prob.set_val("omega_min", 0.0, units="rpm")
    prob.set_val("omega_max", 100.0, units="rpm")
    prob.set_val("control_maxTS", 90.0, units="m/s")
    prob.set_val("tsr_operational", 10.0)
    prob.set_val("control_pitch", 0.0, units="deg")
    prob.set_val("gearbox_efficiency", 0.975)
    prob.set_val("generator_efficiency", 0.975 * np.ones(n_pc))
    prob.set_val("lss_rpm", np.linspace(0.1, 100, n_pc))
    prob.set_val("drivetrainType", "GEARED")

    prob.set_val("Rhub", 1.0, units="m")
    prob.set_val("Rtip", 70.0, units="m")
    prob.set_val("hub_height", 100.0, units="m")
    prob.set_val("precone", 0.0, units="deg")
    prob.set_val("tilt", 0.0, units="deg")
    prob.set_val("yaw", 0.0, units="deg")
    prob.set_val("precurve", np.zeros(n_span), units="m")
    prob.set_val("precurveTip", 0.0, units="m")
    prob.set_val("presweep", np.zeros(n_span), units="m")
    prob.set_val("presweepTip", 0.0, units="m")

    prob.set_val("shearExp", 0.25)
    prob.set_val("nSector", 4)
    prob.set_val("tiploss", True)
    prob.set_val("hubloss", True)
    prob.set_val("wakerotation", True)
    prob.set_val("usecd", True)

    return prob


class TestServo(unittest.TestCase):
    def testGust(self):
        inputs = {}
        outputs = {}
        discrete_inputs = {}
        discrete_outputs = {}

        myobj = rp.GustETM(std=2.5)

        inputs["V_mean"] = 10.0 * np.ones(1)
        inputs["V_hub"] = 15.0 * np.ones(1)
        discrete_inputs["turbulence_class"] = "A"
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        sigma = 0.32 * (0.072 * 8.0 * 3.5 + 10.0)
        expect = 15.0 + 2.5 * sigma
        self.assertEqual(outputs["V_gust"], expect)

        # Test lower case
        discrete_inputs["turbulence_class"] = "c"
        myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        sigma = 0.24 * (0.072 * 8.0 * 3.5 + 10.0)
        expect = 15.0 + 2.5 * sigma
        self.assertEqual(outputs["V_gust"], expect)

        # Test bad class
        discrete_inputs["turbulence_class"] = "d"
        try:
            myobj.compute(inputs, outputs, discrete_inputs, discrete_outputs)
        except ValueError:
            self.assertTrue(True)

    def testStall(self):
        inputs = {}
        outputs = {}

        myobj = rp.NoStallConstraint()

        (n_span, n_aoa, n_Re, n_tab) = NPZFILE["airfoils_cl"].shape

        inputs["airfoils_cl"] = NPZFILE["airfoils_cl"]
        inputs["airfoils_cd"] = NPZFILE["airfoils_cd"]
        inputs["airfoils_cm"] = NPZFILE["airfoils_cm"]
        inputs["airfoils_aoa"] = NPZFILE["airfoils_aoa"]
        inputs["min_s"] = 0.25
        r = NPZFILE["r"]
        r_hub = NPZFILE["r"][0]
        r_tip = NPZFILE["r"][-1]
        inputs["s"] = (r - r_hub) / (r_tip - r_hub)
        inputs["aoa_along_span"] = 7. * np.ones(len(r))
        inputs["stall_margin"] = np.array([ 3. ])

        modeling_options = {}
        modeling_options["WISDEM"] = {}
        modeling_options["WISDEM"]["RotorSE"] = {}
        modeling_options["WISDEM"]["RotorSE"]["n_span"] = n_span
        modeling_options["WISDEM"]["RotorSE"]["n_aoa"] = n_aoa
        modeling_options["WISDEM"]["RotorSE"]["n_Re"] = n_Re
        modeling_options["WISDEM"]["RotorSE"]["n_tab"] = n_tab

        outputs["stall_angle_along_span"] = np.zeros(len(r))
        outputs["no_stall_constraint"] = np.zeros(len(r))

        myobj.compute(inputs, outputs)

        ref_no_stall_constraint = np.array([0.        , 0.        ,
                             0.        , 0.        , 0.        ,
            0.        , 0.        , 0.77009342, 0.80530567, 0.83153911,
            0.85911535, 0.89343367, 0.97975052, 1.07592565, 1.11609355,
            1.13403252, 1.14891246, 1.14568097, 1.14014398, 1.12239983,
            1.07454375, 1.07115146, 1.07115146, 1.07115146, 1.07115146,
            1.07115146, 1.07115146, 1.07115146, 1.07115146, 1.07115146])
        
        ref_stall_angle_along_span = np.array([1.00000000e-06, 6.24217266e+01, 
                                               3.04993295e+00, 3.86108996e+01,
                2.45945626e+01, 1.84737813e+01, 1.35576462e+01, 1.29854375e+01,
                1.24176450e+01, 1.20258925e+01, 1.16398805e+01, 1.11927727e+01,
                1.02066799e+01, 9.29432250e+00, 8.95982239e+00, 8.81808929e+00,
                8.70388331e+00, 8.72843333e+00, 8.77082210e+00, 8.90948100e+00,
                9.30627534e+00, 9.33574791e+00, 9.33574791e+00, 9.33574791e+00,
                9.33574791e+00, 9.33574791e+00, 9.33574791e+00, 9.33574791e+00,
                9.33574791e+00, 9.33574791e+00])

        npt.assert_almost_equal(outputs["no_stall_constraint"], ref_no_stall_constraint)
        npt.assert_almost_equal(outputs["stall_angle_along_span"], ref_stall_angle_along_span)

    def testRegulationTrajectory(self):
        prob = om.Problem(reports=False)

        (n_span, n_aoa, n_Re, n_tab) = NPZFILE["airfoils_cl"].shape
        n_pc = 22

        modeling_options = {}
        modeling_options["WISDEM"] = {}
        modeling_options["WISDEM"]["RotorSE"] = {}
        modeling_options["WISDEM"]["RotorSE"]["n_span"] = n_span
        modeling_options["WISDEM"]["RotorSE"]["n_aoa"] = n_aoa
        modeling_options["WISDEM"]["RotorSE"]["n_Re"] = n_Re
        modeling_options["WISDEM"]["RotorSE"]["n_tab"] = n_tab
        modeling_options["WISDEM"]["RotorSE"]["regulation_reg_III"] = True
        modeling_options["WISDEM"]["RotorSE"]["n_pc"] = n_pc
        modeling_options["WISDEM"]["RotorSE"]["n_pc_spline"] = n_pc

        prob.model.add_subsystem(
            "powercurve", rp.RegulatedPowerCurve(modeling_options=modeling_options), promotes=["*"]
        )
        prob = fillprob(prob, n_pc, n_span)

        # All reg 2: no maxTS, no max rpm, no power limit
        prob["omega_max"] = 1e3
        prob["control_maxTS"] = 1e5
        prob["rated_power"] = 1e16
        prob.run_model()

        grid0 = np.cumsum(np.abs(np.diff(np.cos(np.linspace(-np.pi / 4.0, np.pi / 2.0, n_pc)))))
        grid1 = (grid0 - grid0[0]) / (grid0[-1] - grid0[0])
        V_expect0 = grid1 * (prob["v_max"] - prob["v_min"]) + prob["v_min"]
        V_spline = np.linspace(prob["v_min"], prob["v_max"], n_pc)
        irated = 12

        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        Omega_tsr = V_expect1 * 10 * 60 / 70.0 / 2.0 / np.pi
        npt.assert_equal(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())
        npt.assert_allclose(prob["Omega"], Omega_tsr)
        npt.assert_equal(prob["pitch"], np.zeros(n_pc))
        npt.assert_array_almost_equal(prob["Cp"], prob["Cp_aero"] * 0.975 * 0.975)
        npt.assert_allclose(prob["Cp"], prob["Cp"][0])
        npt.assert_allclose(prob["Cp_aero"], prob["Cp_aero"][0])
        myCp = prob["P"] / (0.5 * 1.225 * V_expect1**3.0 * np.pi * 70**2)
        npt.assert_allclose(myCp, myCp[0], rtol=1e-6)
        self.assertGreater(myCp[0], 0.4)
        self.assertGreater(0.5, myCp[0])
        npt.assert_allclose(myCp, prob["Cp"], rtol=1e-6)
        npt.assert_array_less(prob["P"][:-2], prob["P"][1:-1])
        npt.assert_array_less(prob["Q"][:-2], prob["Q"][1:-1])
        npt.assert_array_less(prob["T"][:-2], prob["T"][1:-1])
        self.assertEqual(prob["rated_V"], V_expect1[-1])
        self.assertAlmostEqual(prob["rated_Omega"][0], Omega_tsr[-1], 5)
        self.assertEqual(prob["rated_pitch"], 0.0)

        # Test no maxTS, max rpm, no power limit
        prob["omega_max"] = 15.0
        prob["control_maxTS"] = 1e5
        prob["rated_power"] = 1e16
        prob.run_model()
        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        Omega_tsr = V_expect1 * 10 * 60 / 70.0 / 2.0 / np.pi
        Omega_expect = np.minimum(Omega_tsr, 15.0)
        npt.assert_allclose(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())
        npt.assert_allclose(prob["Omega"], Omega_expect)
        npt.assert_equal(prob["pitch"][:irated], 0.0)
        npt.assert_array_less(0.0, np.abs(prob["pitch"][(irated + 1) :]))
        npt.assert_array_almost_equal(prob["Cp"], prob["Cp_aero"] * 0.975 * 0.975)
        npt.assert_array_less(prob["P"][:-2], prob["P"][1:-1])
        npt.assert_array_less(prob["Q"][:-2], prob["Q"][1:-1])
        npt.assert_array_less(prob["T"][:-2], prob["T"][1:-1])
        self.assertAlmostEqual(prob["rated_V"], V_expect1[-1], 3)
        self.assertAlmostEqual(prob["rated_Omega"][0], 15.0)
        self.assertEqual(prob["rated_pitch"], 0.0)
        myCp = prob["P"] / (0.5 * 1.225 * V_expect1**3.0 * np.pi * 70**2)
        npt.assert_allclose(myCp[:irated], myCp[0])
        npt.assert_allclose(myCp[:irated], prob["Cp"][:irated])

        # Test maxTS, no max rpm, no power limit
        prob["omega_max"] = 1e3
        prob["control_maxTS"] = 105.0
        prob["rated_power"] = 1e16
        prob.run_model()
        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        # V_expect1[irated] = 105./10.
        Omega_tsr = V_expect1 * 10 * 60 / 70.0 / 2.0 / np.pi
        Omega_expect = np.minimum(Omega_tsr, 105.0 / 70.0 / 2 / np.pi * 60)
        npt.assert_allclose(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())
        npt.assert_allclose(prob["Omega"], Omega_expect)
        npt.assert_equal(prob["pitch"][:irated], 0.0)
        npt.assert_array_less(0.0, np.abs(prob["pitch"][irated:]))
        npt.assert_array_almost_equal(prob["Cp"], prob["Cp_aero"] * 0.975 * 0.975)
        npt.assert_array_less(prob["P"][:-2], prob["P"][1:-1])
        npt.assert_array_less(prob["Q"][:-2], prob["Q"][1:-1])
        npt.assert_array_less(prob["T"][:-2], prob["T"][1:-1])
        self.assertEqual(prob["rated_V"], V_expect1[-1])
        self.assertAlmostEqual(prob["rated_Omega"][0], Omega_expect[-1])
        self.assertEqual(prob["rated_pitch"], 0.0)
        myCp = prob["P"] / (0.5 * 1.225 * V_expect1**3.0 * np.pi * 70**2)
        npt.assert_allclose(myCp[:irated], myCp[0])
        npt.assert_allclose(myCp[:irated], prob["Cp"][:irated])

        # Test no maxTS, no max rpm, power limit
        prob["omega_max"] = 1e3
        prob["control_maxTS"] = 1e4
        prob["rated_power"] = 5e6
        prob.run_model()
        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        Omega_tsr = V_expect1 * 10 * 60 / 70.0 / 2.0 / np.pi
        Omega_expect = np.minimum(Omega_tsr, prob["rated_Omega"])
        npt.assert_allclose(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())
        npt.assert_allclose(prob["Omega"], Omega_expect)
        npt.assert_equal(prob["pitch"][:irated], 0.0)
        npt.assert_array_less(0.0, np.abs(prob["pitch"][(irated + 1) :]))
        npt.assert_array_almost_equal(prob["Cp"], prob["Cp_aero"] * 0.975 * 0.975)
        npt.assert_array_less(prob["P"][:irated], prob["P"][1 : (irated + 1)])
        npt.assert_allclose(prob["P"][irated:], 5e6, rtol=1e-4, atol=0)
        # npt.assert_array_less(prob['Q'], prob['Q'][1:])
        npt.assert_array_less(prob["T"], prob["T"][irated] + 1e-1)
        # print('RATED T',prob["T"][irated])
        # self.assertEqual(prob['rated_V'], V_expect1[-1])
        self.assertAlmostEqual(prob["rated_Omega"][0], Omega_expect[-1])
        self.assertEqual(prob["rated_pitch"], 0.0)
        myCp = prob["P"] / (0.5 * 1.225 * V_expect1**3.0 * np.pi * 70**2)
        npt.assert_allclose(myCp[:irated], myCp[0])
        npt.assert_allclose(myCp[:irated], prob["Cp"][:irated])

        # Test min & max rpm, no power limit
        prob["omega_min"] = 7.0
        prob["omega_max"] = 15.0
        prob["control_maxTS"] = 1e5
        prob["rated_power"] = 1e16
        prob.run_model()
        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        # V_expect1[irated] = 15.*70*2*np.pi/(10.*60.)
        Omega_tsr = V_expect1 * 10 * 60 / 70.0 / 2.0 / np.pi
        Omega_expect = np.maximum(np.minimum(Omega_tsr, 15.0), 7.0)
        npt.assert_allclose(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())
        npt.assert_allclose(prob["Omega"], Omega_expect)
        npt.assert_array_less(0.0, np.abs(prob["pitch"][Omega_expect != Omega_tsr]))
        npt.assert_array_almost_equal(prob["Cp"], prob["Cp_aero"] * 0.975 * 0.975)
        npt.assert_array_less(prob["P"][:-2], prob["P"][1:-1])
        npt.assert_array_less(prob["Q"][:-2], prob["Q"][1:-1])
        npt.assert_array_less(prob["T"][:-2], prob["T"][1:-1])
        self.assertEqual(prob["rated_V"], V_expect1[-1])
        self.assertAlmostEqual(prob["rated_Omega"][0], 15.0)
        self.assertEqual(prob["rated_pitch"], 0.0)
        myCp = prob["P"] / (0.5 * 1.225 * V_expect1**3.0 * np.pi * 70**2)
        npt.assert_allclose(myCp[Omega_expect == Omega_tsr], myCp[6])
        npt.assert_allclose(myCp[Omega_expect == Omega_tsr], prob["Cp"][Omega_expect == Omega_tsr])

        # Test min & max rpm, normal power
        prob["omega_min"] = 7.0
        prob["omega_max"] = 14.0
        prob["control_maxTS"] = 1e5
        prob["rated_power"] = 5e6
        prob.run_model()
        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        # V_expect1[irated] = 14.*70*2*np.pi/(10.*60.)
        Omega_tsr = V_expect1 * 10 * 60 / 70.0 / 2.0 / np.pi
        Omega_expect = np.maximum(np.minimum(Omega_tsr, 14.0), 7.0)
        npt.assert_allclose(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())
        npt.assert_allclose(prob["Omega"], Omega_expect)
        npt.assert_array_less(0.0, np.abs(prob["pitch"][Omega_expect != Omega_tsr]))
        npt.assert_array_almost_equal(prob["Cp"], prob["Cp_aero"] * 0.975 * 0.975)
        npt.assert_array_less(prob["P"][:irated], prob["P"][1 : (irated + 1)])
        npt.assert_allclose(prob["P"][irated:], 5e6, rtol=1e-4, atol=0)
        npt.assert_array_less(prob["T"], prob["T"][irated] + 1e-1)
        self.assertAlmostEqual(prob["rated_Omega"][0], 14.0)
        self.assertLess(0.0, prob["rated_pitch"])
        myCp = prob["P"] / (0.5 * 1.225 * V_expect1**3.0 * np.pi * 70**2)
        npt.assert_allclose(myCp[Omega_expect == Omega_tsr][:-1], myCp[6])
        npt.assert_allclose(myCp[Omega_expect == Omega_tsr][:-1], prob["Cp"][Omega_expect == Omega_tsr][:-1])

        # Test fixed pitch
        prob["omega_min"] = 0.0
        prob["omega_max"] = 15.0
        prob["control_maxTS"] = 1e5
        prob["rated_power"] = 1e16
        prob["control_pitch"] = 5.0
        prob.run_model()
        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        # V_expect1[irated] = 15.*70*2*np.pi/(10.*60.)
        Omega_tsr = V_expect1 * 10 * 60 / 70.0 / 2.0 / np.pi
        Omega_expect = np.minimum(Omega_tsr, 15.0)
        npt.assert_allclose(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())
        npt.assert_allclose(prob["Omega"], Omega_expect)
        npt.assert_equal(prob["pitch"][:irated], 5.0)
        npt.assert_array_less(0.0, np.abs(prob["pitch"][irated:]))
        npt.assert_array_almost_equal(prob["Cp"], prob["Cp_aero"] * 0.975 * 0.975)
        npt.assert_array_less(prob["P"][:-2], prob["P"][1:-1])
        npt.assert_array_less(prob["Q"][:-2], prob["Q"][1:-1])
        npt.assert_array_less(prob["T"][:-2], prob["T"][1:-1])
        self.assertAlmostEqual(prob["rated_V"], V_expect1[-1], 3)
        self.assertAlmostEqual(prob["rated_Omega"][0], 15.0)
        self.assertEqual(prob["rated_pitch"], 5.0)
        myCp = prob["P"] / (0.5 * 1.225 * V_expect1**3.0 * np.pi * 70**2)
        npt.assert_allclose(myCp[:irated], myCp[0])
        npt.assert_allclose(myCp[:irated], prob["Cp"][:irated])

    def testRegulationTrajectoryNoRegion3(self):
        prob = om.Problem(reports=False)

        # Load in airfoil and blade shape inputs for NREL 5MW
        (n_span, n_aoa, n_Re, n_tab) = NPZFILE["airfoils_cl"].shape
        n_pc = 22

        modeling_options = {}
        modeling_options["WISDEM"] = {}
        modeling_options["WISDEM"]["RotorSE"] = {}
        modeling_options["WISDEM"]["RotorSE"]["n_span"] = n_span
        modeling_options["WISDEM"]["RotorSE"]["n_aoa"] = n_aoa
        modeling_options["WISDEM"]["RotorSE"]["n_Re"] = n_Re
        modeling_options["WISDEM"]["RotorSE"]["n_tab"] = n_tab
        modeling_options["WISDEM"]["RotorSE"]["regulation_reg_III"] = False
        modeling_options["WISDEM"]["RotorSE"]["n_pc"] = n_pc
        modeling_options["WISDEM"]["RotorSE"]["n_pc_spline"] = n_pc

        prob.model.add_subsystem(
            "powercurve", rp.RegulatedPowerCurve(modeling_options=modeling_options), promotes=["*"]
        )
        prob = fillprob(prob, n_pc, n_span)

        # All reg 2: no maxTS, no max rpm, no power limit
        prob["omega_max"] = 1e3
        prob["control_maxTS"] = 1e5
        prob["rated_power"] = 1e16
        prob.run_model()

        grid0 = np.cumsum(np.abs(np.diff(np.cos(np.linspace(-np.pi / 4.0, np.pi / 2.0, n_pc)))))
        grid1 = (grid0 - grid0[0]) / (grid0[-1] - grid0[0])
        V_expect0 = grid1 * (prob["v_max"] - prob["v_min"]) + prob["v_min"]
        V_spline = np.linspace(prob["v_min"], prob["v_max"], n_pc)
        irated = 12

        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        Omega_tsr = V_expect1 * 10 * 60 / 70.0 / 2.0 / np.pi
        npt.assert_equal(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())
        npt.assert_allclose(prob["Omega"], Omega_tsr)
        npt.assert_equal(prob["pitch"], np.zeros(n_pc))
        npt.assert_array_almost_equal(prob["Cp"], prob["Cp_aero"] * 0.975 * 0.975)
        npt.assert_allclose(prob["Cp"], prob["Cp"][0], rtol=1e-6)
        npt.assert_allclose(prob["Cp_aero"], prob["Cp_aero"][0])
        myCp = prob["P"] / (0.5 * 1.225 * V_expect1**3.0 * np.pi * 70**2)
        npt.assert_allclose(myCp, myCp[0], rtol=1e-6)
        self.assertGreater(myCp[0], 0.4)
        self.assertGreater(0.5, myCp[0])
        npt.assert_allclose(myCp, prob["Cp"], rtol=1e-6)
        npt.assert_array_less(prob["P"][:-2], prob["P"][1:-1])
        npt.assert_array_less(prob["Q"][:-2], prob["Q"][1:-1])
        npt.assert_array_less(prob["T"][:-2], prob["T"][1:-1])
        self.assertEqual(prob["rated_V"], V_expect1[-1])
        self.assertAlmostEqual(prob["rated_Omega"][0], Omega_tsr[-1], 5)
        self.assertEqual(prob["rated_pitch"], 0.0)

        # Test no maxTS, no max rpm, power limit
        prob["omega_max"] = 1e3
        prob["control_maxTS"] = 1e4
        prob["rated_power"] = 5e6
        prob.run_model()
        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        Omega_tsr = V_expect1 * 10 * 60 / 70.0 / 2.0 / np.pi
        Omega_expect = np.minimum(Omega_tsr, prob["rated_Omega"])
        npt.assert_allclose(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())
        npt.assert_allclose(prob["Omega"], Omega_expect)
        npt.assert_equal(prob["pitch"], 0.0)
        # npt.assert_array_less(0.0, np.abs(prob['pitch'][(irated+1):]))
        npt.assert_allclose(prob["Cp"][: (irated + 1)], prob["Cp_aero"][: (irated + 1)] * 0.975 * 0.975)
        npt.assert_array_less(prob["P"][:irated], prob["P"][1 : (irated + 1)])
        npt.assert_allclose(prob["P"][irated:], 5e6, rtol=1e-6, atol=0)
        # npt.assert_equal(prob['Q'][(irated+1):], prob['Q'][irated])
        npt.assert_equal(prob["T"][(irated + 1) :], 0.0)
        npt.assert_array_less(prob["T"], prob["T"][irated] + 1e-1)
        # self.assertEqual(prob['rated_V'], V_expect1[-1])
        self.assertAlmostEqual(prob["rated_Omega"][0], Omega_expect[-1])
        self.assertEqual(prob["rated_pitch"], 0.0)
        myCp = prob["P"] / (0.5 * 1.225 * V_expect1**3.0 * np.pi * 70**2)
        npt.assert_allclose(myCp[:irated], myCp[0])
        npt.assert_allclose(myCp[:irated], prob["Cp"][:irated])

    def testRegulationTrajectory_PeakShaving(self):
        prob = om.Problem(reports=False)

        (n_span, n_aoa, n_Re, n_tab) = NPZFILE["airfoils_cl"].shape
        n_pc = 22

        modeling_options = {}
        modeling_options["WISDEM"] = {}
        modeling_options["WISDEM"]["RotorSE"] = {}
        modeling_options["WISDEM"]["RotorSE"]["n_span"] = n_span
        modeling_options["WISDEM"]["RotorSE"]["n_aoa"] = n_aoa
        modeling_options["WISDEM"]["RotorSE"]["n_Re"] = n_Re
        modeling_options["WISDEM"]["RotorSE"]["n_tab"] = n_tab
        modeling_options["WISDEM"]["RotorSE"]["regulation_reg_III"] = True
        modeling_options["WISDEM"]["RotorSE"]["n_pc"] = n_pc
        modeling_options["WISDEM"]["RotorSE"]["n_pc_spline"] = n_pc

        prob.model.add_subsystem(
            "powercurve", rp.RegulatedPowerCurve(modeling_options=modeling_options), promotes=["*"]
        )
        prob = fillprob(prob, n_pc, n_span)

        # All reg 2: no maxTS, no max rpm, no power limit
        prob["omega_max"] = 1e3
        prob["control_maxTS"] = 1e5
        prob["rated_power"] = 1e16
        prob["ps_percent"] = 0.8
        prob.run_model()

        grid0 = np.cumsum(np.abs(np.diff(np.cos(np.linspace(-np.pi / 4.0, np.pi / 2.0, n_pc)))))
        grid1 = (grid0 - grid0[0]) / (grid0[-1] - grid0[0])
        V_expect0 = grid1 * (prob["v_max"] - prob["v_min"]) + prob["v_min"]
        V_spline = np.linspace(prob["v_min"], prob["v_max"], n_pc)
        irated = 12

        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        Omega_tsr = V_expect1 * 10 * 60 / 70.0 / 2.0 / np.pi
        npt.assert_equal(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())
        npt.assert_allclose(prob["Omega"], Omega_tsr)
        npt.assert_equal(prob["pitch"], np.zeros(n_pc))
        npt.assert_array_almost_equal(prob["Cp"], prob["Cp_aero"] * 0.975 * 0.975)
        npt.assert_allclose(prob["Cp"], prob["Cp"][0], rtol=1e-6)
        npt.assert_allclose(prob["Cp_aero"], prob["Cp_aero"][0])
        myCp = prob["P"] / (0.5 * 1.225 * V_expect1**3.0 * np.pi * 70**2)
        npt.assert_allclose(myCp, myCp[0], rtol=1e-6)
        self.assertGreater(myCp[0], 0.4)
        self.assertGreater(0.5, myCp[0])
        npt.assert_allclose(myCp, prob["Cp"], rtol=1e-6)
        npt.assert_array_less(prob["P"][:-2], prob["P"][1:-1])
        npt.assert_array_less(prob["Q"][:-2], prob["Q"][1:-1])
        npt.assert_array_less(prob["T"][:-2], prob["T"][1:-1])
        self.assertEqual(prob["rated_V"], V_expect1[-1])
        self.assertAlmostEqual(prob["rated_Omega"][0], Omega_tsr[-1], 5)
        self.assertEqual(prob["rated_pitch"], 0.0)

        # Test no maxTS, max rpm, no power limit
        prob["omega_max"] = 15.0
        prob["control_maxTS"] = 1e5
        prob["rated_power"] = 1e16
        prob.run_model()
        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        Omega_tsr = V_expect1 * 10 * 60 / 70.0 / 2.0 / np.pi
        Omega_expect = np.minimum(Omega_tsr, 15.0)
        npt.assert_allclose(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())
        npt.assert_allclose(prob["Omega"], Omega_expect)
        npt.assert_equal(prob["pitch"][:irated], 0.0)
        npt.assert_array_less(0.0, np.abs(prob["pitch"][(irated + 1) :]))
        npt.assert_array_almost_equal(prob["Cp"], prob["Cp_aero"] * 0.975 * 0.975)
        npt.assert_array_less(prob["P"][:-2], prob["P"][1:-1])
        npt.assert_array_less(prob["Q"][:-2], prob["Q"][1:-1])
        npt.assert_array_less(prob["T"][:-2], prob["T"][1:-1])
        self.assertAlmostEqual(prob["rated_V"], V_expect1[-1], 3)
        self.assertAlmostEqual(prob["rated_Omega"][0], 15.0)
        self.assertEqual(prob["rated_pitch"], 0.0)
        myCp = prob["P"] / (0.5 * 1.225 * V_expect1**3.0 * np.pi * 70**2)
        npt.assert_allclose(myCp[:irated], myCp[0])
        npt.assert_allclose(myCp[:irated], prob["Cp"][:irated])

        # Test maxTS, no max rpm, no power limit
        prob["omega_max"] = 1e3
        prob["control_maxTS"] = 105.0
        prob["rated_power"] = 1e16
        prob.run_model()
        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        # V_expect1[irated] = 105./10.
        Omega_tsr = V_expect1 * 10 * 60 / 70.0 / 2.0 / np.pi
        Omega_expect = np.minimum(Omega_tsr, 105.0 / 70.0 / 2 / np.pi * 60)
        npt.assert_allclose(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())
        npt.assert_allclose(prob["Omega"], Omega_expect)
        npt.assert_equal(prob["pitch"][:irated], 0.0)
        npt.assert_array_less(0.0, np.abs(prob["pitch"][irated:]))
        npt.assert_array_almost_equal(prob["Cp"], prob["Cp_aero"] * 0.975 * 0.975)
        npt.assert_array_less(prob["P"][:-2], prob["P"][1:-1])
        npt.assert_array_less(prob["Q"][:-2], prob["Q"][1:-1])
        npt.assert_array_less(prob["T"][:-2], prob["T"][1:-1])
        self.assertEqual(prob["rated_V"], V_expect1[-1])
        self.assertAlmostEqual(prob["rated_Omega"][0], Omega_expect[-1])
        self.assertEqual(prob["rated_pitch"], 0.0)
        myCp = prob["P"] / (0.5 * 1.225 * V_expect1**3.0 * np.pi * 70**2)
        npt.assert_allclose(myCp[:irated], myCp[0])
        npt.assert_allclose(myCp[:irated], prob["Cp"][:irated])

        # Test no maxTS, no max rpm, power limit
        prob["omega_max"] = 1e3
        prob["control_maxTS"] = 1e4
        prob["rated_power"] = 5e6
        prob.run_model()
        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        Omega_tsr = V_expect1 * 10 * 60 / 70.0 / 2.0 / np.pi
        Omega_expect = np.minimum(Omega_tsr, prob["rated_Omega"])
        npt.assert_allclose(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())
        npt.assert_allclose(prob["Omega"][:irated], Omega_expect[:irated])
        npt.assert_equal(prob["pitch"][: (irated - 1)], 0.0)
        npt.assert_array_less(0.0, np.abs(prob["pitch"][irated:]))
        npt.assert_array_almost_equal(prob["Cp"], prob["Cp_aero"] * 0.975 * 0.975)
        npt.assert_array_less(prob["P"][:irated], prob["P"][1 : (irated + 1)])
        npt.assert_allclose(prob["P"][irated:], 5e6, rtol=1e-4, atol=0)
        npt.assert_array_less(prob["T"], 0.8 * 880899)  # From print out in first test
        self.assertAlmostEqual(prob["rated_Omega"][0], Omega_expect[-1])
        self.assertGreater(prob["rated_pitch"], 0.0)
        myCp = prob["P"] / (0.5 * 1.225 * V_expect1**3.0 * np.pi * 70**2)
        npt.assert_allclose(myCp[: (irated - 1)], myCp[0])
        npt.assert_allclose(myCp[: (irated - 1)], prob["Cp"][: (irated - 1)])

        # Test min & max rpm, no power limit
        prob["omega_min"] = 7.0
        prob["omega_max"] = 15.0
        prob["control_maxTS"] = 1e5
        prob["rated_power"] = 1e16
        prob.run_model()
        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        # V_expect1[irated] = 15.*70*2*np.pi/(10.*60.)
        Omega_tsr = V_expect1 * 10 * 60 / 70.0 / 2.0 / np.pi
        Omega_expect = np.maximum(np.minimum(Omega_tsr, 15.0), 7.0)
        npt.assert_allclose(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())
        npt.assert_allclose(prob["Omega"], Omega_expect)
        npt.assert_array_less(0.0, np.abs(prob["pitch"][Omega_expect != Omega_tsr]))
        npt.assert_array_almost_equal(prob["Cp"], prob["Cp_aero"] * 0.975 * 0.975)
        npt.assert_array_less(prob["P"][:-2], prob["P"][1:-1])
        npt.assert_array_less(prob["Q"][:-2], prob["Q"][1:-1])
        npt.assert_array_less(prob["T"][:-2], prob["T"][1:-1])
        self.assertEqual(prob["rated_V"], V_expect1[-1])
        self.assertAlmostEqual(prob["rated_Omega"][0], 15.0)
        self.assertEqual(prob["rated_pitch"], 0.0)
        myCp = prob["P"] / (0.5 * 1.225 * V_expect1**3.0 * np.pi * 70**2)
        npt.assert_allclose(myCp[Omega_expect == Omega_tsr], myCp[6])
        npt.assert_allclose(myCp[Omega_expect == Omega_tsr], prob["Cp"][Omega_expect == Omega_tsr])

        # Test min & max rpm, normal power
        prob["omega_min"] = 7.0
        prob["omega_max"] = 14.0
        prob["control_maxTS"] = 1e5
        prob["rated_power"] = 5e6
        prob.run_model()
        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        # V_expect1[irated] = 14.*70*2*np.pi/(10.*60.)
        Omega_tsr = V_expect1 * 10 * 60 / 70.0 / 2.0 / np.pi
        Omega_expect = np.maximum(np.minimum(Omega_tsr, 14.0), 7.0)
        npt.assert_allclose(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())
        npt.assert_allclose(prob["Omega"], Omega_expect)
        npt.assert_array_less(0.0, np.abs(prob["pitch"][Omega_expect != Omega_tsr]))
        npt.assert_array_almost_equal(prob["Cp"], prob["Cp_aero"] * 0.975 * 0.975)
        npt.assert_array_less(prob["P"][:irated], prob["P"][1 : (irated + 1)])
        npt.assert_allclose(prob["P"][irated:], 5e6, rtol=1e-4, atol=0)
        npt.assert_array_less(prob["T"], 0.8 * 880899)  # From print out in first test
        self.assertAlmostEqual(prob["rated_Omega"][0], 14.0)
        self.assertLess(0.0, prob["rated_pitch"])
        myCp = prob["P"] / (0.5 * 1.225 * V_expect1**3.0 * np.pi * 70**2)
        npt.assert_allclose(myCp[Omega_expect == Omega_tsr][:-1], myCp[6])
        npt.assert_allclose(myCp[Omega_expect == Omega_tsr][:-1], prob["Cp"][Omega_expect == Omega_tsr][:-1])

    def testRegulationTrajectory_reindex(self):
        prob = om.Problem(reports=False)

        debug_archive = os.path.dirname(os.path.abspath(__file__)) + os.path.sep + "debug.npz"
        debug_npz = np.load(debug_archive)

        (n_span, n_aoa, n_Re, n_tab) = debug_npz["airfoils_cl"].shape
        n_pc = 50

        modeling_options = {}
        modeling_options["WISDEM"] = {}
        modeling_options["WISDEM"]["RotorSE"] = {}
        modeling_options["WISDEM"]["RotorSE"]["n_span"] = n_span
        modeling_options["WISDEM"]["RotorSE"]["n_aoa"] = n_aoa
        modeling_options["WISDEM"]["RotorSE"]["n_Re"] = n_Re
        modeling_options["WISDEM"]["RotorSE"]["n_tab"] = n_tab
        modeling_options["WISDEM"]["RotorSE"]["regulation_reg_III"] = True
        modeling_options["WISDEM"]["RotorSE"]["n_pc"] = n_pc
        modeling_options["WISDEM"]["RotorSE"]["n_pc_spline"] = n_pc

        prob.model.add_subsystem(
            "powercurve", rp.RegulatedPowerCurve(modeling_options=modeling_options), promotes=["*"]
        )
        prob.setup()
        for k in debug_npz.files:
            prob[k] = debug_npz[k]

        prob.run_model()
        grid0 = np.cumsum(np.abs(np.diff(np.cos(np.linspace(-np.pi / 4.0, np.pi / 2.0, n_pc)))))
        grid1 = (grid0 - grid0[0]) / (grid0[-1] - grid0[0])
        V_expect0 = grid1 * (prob["v_max"] - prob["v_min"]) + prob["v_min"]
        V_spline = np.linspace(prob["v_min"], prob["v_max"], n_pc)

        V_expect1 = np.sort(np.r_[V_expect0, prob["rated_V"]])
        npt.assert_equal(prob["V"], V_expect1)
        npt.assert_equal(prob["V_spline"], V_spline.flatten())


def suite():
    suite = [
        unittest.TestLoader().loadTestsFromTestCase(TestServo),
    ]
    return unittest.TestSuite(suite)


if __name__ == "__main__":
    result = unittest.TextTestRunner().run(suite())

    if result.wasSuccessful():
        exit(0)
    else:
        exit(1)



import os
import sys
import logging
import warnings
import time

import numpy as np
import openmdao.api as om
from openmdao.utils.mpi import MPI
if MPI:
    max_cores = MPI.COMM_WORLD.Get_size()
    
from wisdem.commonse import fileIO
from wisdem.glue_code.glue_code import WindPark
from wisdem.glue_code.gc_LoadInputs import WindTurbineOntologyPython
from wisdem.glue_code.gc_WT_InitModel import yaml2openmdao
from wisdem.glue_code.gc_PoseOptimization import PoseOptimization

# Numpy deprecation warnings
warnings.filterwarnings("ignore", category=np.exceptions.VisibleDeprecationWarning)

# Suppress the maxfev warnings is scipy _minpack_py, line:175
warnings.simplefilter("ignore", RuntimeWarning, lineno=175)
warnings.simplefilter("ignore", RuntimeWarning, lineno=177)

def run_wisdem(fname_wt_input, fname_modeling_options, fname_opt_options, overridden_values=None, run_only=False):
    t0 = time.time()

    # Load all yaml inputs and validate (also fills in defaults)
    wt_initial = WindTurbineOntologyPython(fname_wt_input, fname_modeling_options, fname_opt_options)
    wt_init, modeling_options, opt_options = wt_initial.get_input_data()

    myopt = PoseOptimization(wt_init, modeling_options, opt_options)

    folder_output = opt_options["general"]["folder_output"]

    os.makedirs(folder_output, exist_ok=True)

    # create logger
    logger = logging.getLogger("wisdem/weis")
    logger.setLevel(logging.INFO)

    # create handlers
    ht = logging.StreamHandler()
    ht.setLevel(logging.WARNING)

    flog = os.path.join(folder_output, opt_options["general"]["fname_output"] + ".log")
    hf = logging.FileHandler(flog, mode="w")
    hf.setLevel(logging.INFO)

    # create formatters
    formatter_t = logging.Formatter("%(module)s:%(funcName)s:%(lineno)d %(levelname)s:%(message)s")
    formatter_f = logging.Formatter(
        "P%(process)d %(asctime)s %(module)s:%(funcName)s:%(lineno)d %(levelname)s:%(message)s"
    )

    # add formatter to handlers
    ht.setFormatter(formatter_t)
    hf.setFormatter(formatter_f)

    # add handlers to logger
    logger.addHandler(ht)
    logger.addHandler(hf)
    logger.info("Started")

    if MPI and opt_options["opt_flag"] and not run_only:
        # Parallel settings for OpenMDAO
        wt_opt = om.Problem(model=om.Group(num_par_fd=max_cores), reports=False)
        wt_opt.model.add_subsystem(
            "comp", WindPark(modeling_options=modeling_options, opt_options=opt_options), promotes=["*"]
        )
    else:
        # Sequential finite differencing
        wt_opt = om.Problem(
            model=WindPark(modeling_options=modeling_options, opt_options=opt_options), reports=False
        )

    # If at least one of the design variables is active, setup an optimization
    if opt_options["opt_flag"] and not run_only:
        wt_opt = myopt.set_driver(wt_opt)
        wt_opt = myopt.set_objective(wt_opt)
        wt_opt = myopt.set_design_variables(wt_opt, wt_init)
        wt_opt = myopt.set_constraints(wt_opt)
        wt_opt = myopt.set_recorders(wt_opt)

    if modeling_options["General"]["verbosity"] == False:
        wt_opt.set_solver_print(level=-1)

    # Set working directory and setup openmdao problem
    wt_opt.options['work_dir'] = folder_output
    wt_opt.setup()

    # Load initial wind turbine data from wt_initial to the openmdao problem
    wt_opt = yaml2openmdao(wt_opt, modeling_options, wt_init, opt_options)
    wt_opt = myopt.set_initial(wt_opt, wt_init)

    # If the user provides values in this dict, they overwrite
    # whatever values have been set by the yaml files.
    # This is useful for performing black-box wrapped optimization without
    # needing to modify the yaml files.
    if overridden_values is not None:
        for key in overridden_values:
            wt_opt[key] = overridden_values[key]

    # Place the last design variables from a previous run into the problem.
    # This needs to occur after the above setup() and yaml2openmdao() calls
    # so these values are correctly placed in the problem.
    wt_opt = myopt.set_restart(wt_opt)

    if "check_totals" in opt_options["driver"] and not run_only:
        if opt_options["driver"]["check_totals"]:
            wt_opt.run_model()
            totals = wt_opt.compute_totals()

    if "check_partials" in opt_options["driver"] and not run_only:
        if opt_options["driver"]["check_partials"]:
            wt_opt.run_model()
            checks = wt_opt.check_partials(compact_print=True)

    sys.stdout.flush()

    if opt_options["driver"]["step_size_study"]["flag"] and not run_only:
        wt_opt.run_model()
        study_options = opt_options["driver"]["step_size_study"]
        step_sizes = study_options["step_sizes"]
        all_derivs = {}
        for idx, step_size in enumerate(step_sizes):
            wt_opt.model.approx_totals(method="fd", step=step_size, form=study_options["form"])

            if study_options["of"]:
                of = study_options["of"]
            else:
                of = None

            if study_options["wrt"]:
                wrt = study_options["wrt"]
            else:
                wrt = None

            derivs = wt_opt.compute_totals(of=of, wrt=wrt, driver_scaling=study_options["driver_scaling"])
            all_derivs[idx] = derivs
            all_derivs[idx]["step_size"] = step_size
        np.save("total_derivs.npy", all_derivs)

    # Run openmdao problem
    elif opt_options["opt_flag"] and not run_only:
        wt_opt.run_driver()
    else:
        wt_opt.run_model()

    # Save data coming from openmdao to an output yaml file
    froot_out = os.path.join(folder_output, opt_options["general"]["fname_output"])
    wt_initial.write_ontology(wt_opt, froot_out)
    wt_initial.write_options(froot_out)

    # Save data to numpy and matlab arrays
    fileIO.save_data(froot_out, wt_opt)

    t1 = time.time()
    if MPI:
        rank = MPI.COMM_WORLD.Get_rank()
    else:
        rank = 0
    if rank == 0:
        print("WISDEM run completed in,", t1-t0, "seconds")

    return wt_opt, modeling_options, opt_options


def load_wisdem(frootin):
    froot,fext = os.path.splitext(frootin)
    if fext not in ['.yaml','.pkl']:
        froot = frootin
    fgeom = froot + ".yaml"
    fmodel = froot + "-modeling.yaml"
    fopt = froot + "-analysis.yaml"
    fpkl = froot + ".pkl"

    # Load all yaml inputs and validate (also fills in defaults)
    wt_initial = WindTurbineOntologyPython(fgeom, fmodel, fopt)
    wt_init, modeling_options, opt_options = wt_initial.get_input_data()

    wt_opt = om.Problem(model=WindPark(modeling_options=modeling_options, opt_options=opt_options), reports=False)
    wt_opt.setup()

    wt_opt = fileIO.load_data(fpkl, wt_opt)

    return wt_opt, modeling_options, opt_options
